# Original work Copyright Fabio Zadrozny (EPL 1.0)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
    This module provides utilities to get the absolute filenames so that we can be sure that:
        - The case of a file will match the actual file in the filesystem (otherwise breakpoints won't be hit).
        - Providing means for the user to make path conversions when doing a remote debugging session in
          one machine and debugging in another.

    To do that, the PATHS_FROM_IDE_TO_PYTHON constant must be filled with the appropriate paths.

    @note:
        in this context, the server is where your python process is running
        and the client is where eclipse is running.

    E.g.:
        If the server (your python process) has the structure
            /user/projects/my_project/src/package/module1.py

        and the client has:
            c:\my_project\src\package\module1.py

        the PATHS_FROM_IDE_TO_PYTHON would have to be:
            PATHS_FROM_IDE_TO_PYTHON = [(r'c:\my_project\src', r'/user/projects/my_project/src')]

        alternatively, this can be set with an environment variable from the command line:
           set PATHS_FROM_IDE_TO_PYTHON=[['c:\my_project\src','/user/projects/my_project/src']]

    @note: DEBUG_CLIENT_SERVER_TRANSLATION can be set to True to debug the result of those translations

    @note: the case of the paths is important! Note that this can be tricky to get right when one machine
    uses a case-independent filesystem and the other uses a case-dependent filesystem (if the system being
    debugged is case-independent, 'normcase()' should be used on the paths defined in PATHS_FROM_IDE_TO_PYTHON).

    @note: all the paths with breakpoints must be translated (otherwise they won't be found in the server)

    @note: to enable remote debugging in the target machine (pydev extensions in the eclipse installation)
        import pydevd;pydevd.settrace(host, stdoutToServer, stderrToServer, port, suspend)

        see parameter docs on pydevd.py

    @note: for doing a remote debugging session, all the pydevd_ files must be on the server accessible
        through the PYTHONPATH (and the PATHS_FROM_IDE_TO_PYTHON only needs to be set on the target
        machine for the paths that'll actually have breakpoints).
"""


import json
import os.path
import sys
import itertools
import ntpath
from robocorp_ls_core.constants import IS_WINDOWS, IS_JYTHON
from robocorp_ls_core.robotframework_log import get_logger, get_log_level
from functools import partial


log = get_logger(__name__)
_nt_os_normcase = ntpath.normcase
basename = os.path.basename
os_path_exists = os.path.exists
join = os.path.join


def __getfilesystemencoding():
    """
    Note: there's a copy of this method in interpreterInfo.py
    """
    try:
        ret = sys.getfilesystemencoding()
        if not ret:
            raise RuntimeError("Unable to get encoding.")
        return ret
    except:
        if sys.platform == "win32":
            return "mbcs"
        return "utf-8"


def getfilesystemencoding():
    try:
        ret = __getfilesystemencoding()

        # Check if the encoding is actually there to be used!
        if hasattr("", "encode"):
            "".encode(ret)
        if hasattr("", "decode"):
            "".decode(ret)

        return ret
    except:
        return "utf-8"


file_system_encoding = getfilesystemencoding()
filesystem_encoding_is_utf8 = file_system_encoding.lower() in ("utf-8", "utf_8", "utf8")

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError  # noqa

rPath = os.path.realpath


def _get_library_dir():
    library_dir = None
    try:
        import sysconfig

        library_dir = sysconfig.get_path("purelib")
    except ImportError:
        pass  # i.e.: Only 2.7 onwards

    if library_dir is None or not os_path_exists(library_dir):
        for path in sys.path:
            if os_path_exists(path) and os.path.basename(path) == "site-packages":
                library_dir = path
                break

    if library_dir is None or not os_path_exists(library_dir):
        library_dir = os.path.dirname(os.__file__)

    return library_dir


# Note: we can't call sysconfig.get_path from _NormPath (it deadlocks on Python 2.7) so, we
# need to get the library dir during module loading.
_library_dir = _get_library_dir()

# defined as a list of tuples where the 1st element of the tuple is the path in the client machine
# and the 2nd element is the path in the server machine.
# see module docstring for more details.
try:
    PATHS_FROM_IDE_TO_PYTHON = json.loads(
        os.environ.get("PATHS_FROM_IDE_TO_PYTHON", "[]")
    )
except Exception:
    log.critical("Error loading PATHS_FROM_IDE_TO_PYTHON from environment variable.")
    log.exception()
    PATHS_FROM_IDE_TO_PYTHON = []
else:
    if not isinstance(PATHS_FROM_IDE_TO_PYTHON, list):
        log.critical(
            "Expected PATHS_FROM_IDE_TO_PYTHON loaded from environment variable to be a list."
        )
        PATHS_FROM_IDE_TO_PYTHON = []
    else:
        # Converting json lists to tuple
        PATHS_FROM_IDE_TO_PYTHON = [tuple(x) for x in PATHS_FROM_IDE_TO_PYTHON]

# example:
# PATHS_FROM_IDE_TO_PYTHON = [
#  (r'd:\temp\temp_workspace_2\test_python\src\yyy\yyy',
#   r'd:\temp\temp_workspace_2\test_python\src\hhh\xxx')
# ]

convert_to_long_pathname = lambda filename: filename
convert_to_short_pathname = lambda filename: filename
get_path_with_real_case = lambda filename: filename

if sys.platform == "win32":
    try:
        import ctypes
        from ctypes.wintypes import MAX_PATH, LPCWSTR, LPWSTR, DWORD

        GetLongPathName = ctypes.windll.kernel32.GetLongPathNameW  # noqa
        GetLongPathName.argtypes = [LPCWSTR, LPWSTR, DWORD]
        GetLongPathName.restype = DWORD

        GetShortPathName = ctypes.windll.kernel32.GetShortPathNameW  # noqa
        GetShortPathName.argtypes = [LPCWSTR, LPWSTR, DWORD]
        GetShortPathName.restype = DWORD

        def _convert_to_long_pathname(filename):
            buf = ctypes.create_unicode_buffer(MAX_PATH)

            rv = GetLongPathName(filename, buf, MAX_PATH)
            if rv != 0 and rv <= MAX_PATH:
                filename = buf.value

            return filename

        def _convert_to_short_pathname(filename):
            buf = ctypes.create_unicode_buffer(MAX_PATH)

            rv = GetShortPathName(filename, buf, MAX_PATH)
            if rv != 0 and rv <= MAX_PATH:
                filename = buf.value

            return filename

        # Note that we have a cache for previous list dirs... the only case where this may be an
        # issue is if the user actually changes the case of an existing file on windows while
        # the debugger is executing (as this seems very unlikely and the cache can save a
        # reasonable time -- especially on mapped drives -- it seems nice to have it).
        _listdir_cache: dict = {}

        def _resolve_listing(resolved, iter_parts, cache=_listdir_cache):
            while True:  # Note: while True to make iterative and not recursive
                try:
                    resolve_lowercase = next(iter_parts)  # must be lowercase already
                except StopIteration:
                    return resolved

                resolved_lower = resolved.lower()

                resolved_joined = cache.get((resolved_lower, resolve_lowercase))
                if resolved_joined is None:
                    dir_contents = cache.get(resolved_lower)
                    if dir_contents is None:
                        dir_contents = cache[resolved_lower] = os.listdir(resolved)

                    for filename in dir_contents:
                        if filename.lower() == resolve_lowercase:
                            resolved_joined = os.path.join(resolved, filename)
                            cache[(resolved_lower, resolve_lowercase)] = resolved_joined
                            break
                    else:
                        raise FileNotFoundError(
                            "Unable to find: %s in %s" % (resolve_lowercase, resolved)
                        )

                resolved = resolved_joined

        def _get_path_with_real_case(filename):
            # Note: this previously made:
            # convert_to_long_pathname(convert_to_short_pathname(filename))
            # but this is no longer done because we can't rely on getting the shortname
            # consistently (there are settings to disable it on Windows).
            # So, using approach which resolves by listing the dir.

            if "~" in filename:
                filename = convert_to_long_pathname(filename)

            if filename.startswith("<") or not os_path_exists(filename):
                return filename  # Not much we can do.

            drive, parts = os.path.splitdrive(os.path.normpath(filename))
            drive = drive.upper()
            while parts.startswith(os.path.sep):
                parts = parts[1:]
                drive += os.path.sep
            parts = parts.lower().split(os.path.sep)

            try:
                return _resolve_listing(drive, iter(parts))
            except FileNotFoundError:
                _listdir_cache.clear()
                # Retry once after clearing the cache we have.
                try:
                    return _resolve_listing(drive, iter(parts))
                except FileNotFoundError:
                    if os_path_exists(filename):
                        # This is really strange, ask the user to report as error.
                        log.critical(
                            "robot debugger: critical: unable to get real case for file. Details:\n"
                            "filename: %s\ndrive: %s\nparts: %s\n"
                            "(please create a ticket in the tracker to address this).",
                            filename,
                            drive,
                            parts,
                        )
                        log.exception()
                    # Don't fail, just return the original file passed.
                    return filename

        # Check that it actually works
        _get_path_with_real_case(__file__)
    except:
        # Something didn't quite work out, leave no-op conversions in place.
        if get_log_level() > 2:
            log.exception()
    else:
        convert_to_long_pathname = _convert_to_long_pathname
        convert_to_short_pathname = _convert_to_short_pathname
        get_path_with_real_case = _get_path_with_real_case


def _normcase_windows(filename):
    # `normcase` doesn't lower case on Python 2 for non-English locale, so we should do it manually.
    if "~" in filename:
        filename = convert_to_long_pathname(filename)

    filename = _nt_os_normcase(filename)
    return filename.lower()


def _normcase_linux(filename):
    return filename  # no-op


if IS_WINDOWS:
    normcase = _normcase_windows

else:
    normcase = _normcase_linux

_ide_os = "WINDOWS" if IS_WINDOWS else "UNIX"

_normcase_from_client = normcase

DEBUG_CLIENT_SERVER_TRANSLATION = os.environ.get(
    "DEBUG_PATHS_TRANSLATION", "False"
).lower() in ("1", "true")


def set_ide_os(os):
    """
    We need to set the IDE os because the host where the code is running may be
    actually different from the client (and the point is that we want the proper
    paths to translate from the client to the server).

    :param os:
        'UNIX' or 'WINDOWS'
    """
    global _ide_os
    global _normcase_from_client
    prev = _ide_os
    if (
        os == "WIN"
    ):  # Apparently PyCharm uses 'WIN' (https://github.com/fabioz/PyDev.Debugger/issues/116)
        os = "WINDOWS"

    assert os in ("WINDOWS", "UNIX")

    if DEBUG_CLIENT_SERVER_TRANSLATION:
        print("robot debugger: client OS: %s" % (os,))

    _normcase_from_client = normcase
    if os == "WINDOWS":

        # Client in Windows and server in Unix, we need to normalize the case.
        if not IS_WINDOWS:
            _normcase_from_client = _normcase_windows

    else:
        # Client in Unix and server in Windows, we can't normalize the case.
        if IS_WINDOWS:
            _normcase_from_client = _normcase_linux

    if prev != os:
        _ide_os = os
        # We need to (re)setup how the client <-> server translation works to provide proper separators.
        setup_client_server_paths(_last_client_server_paths_set)


# Caches filled as requested during the debug session.
NORM_PATHS_CONTAINER: dict = {}
NORM_PATHS_AND_BASE_CONTAINER: dict = {}


def _NormFile(filename):
    _abs_path, real_path = _NormPaths(filename)
    return real_path


def _AbsFile(filename):
    abs_path, _real_path = _NormPaths(filename)
    return abs_path


# Returns tuple of absolute path and real path for given filename
def _NormPaths(filename, NORM_PATHS_CONTAINER=NORM_PATHS_CONTAINER):
    try:
        return NORM_PATHS_CONTAINER[filename]
    except:
        if filename.__class__ != str:
            raise AssertionError(
                "Paths passed to _NormPaths must be str. Found: %s (%s)"
                % (filename, type(filename))
            )
        if os is None:  # Interpreter shutdown
            return filename, filename

        os_path = os.path
        if os_path is None:  # Interpreter shutdown
            return filename, filename

        os_path_abspath = os_path.abspath
        os_path_isabs = os_path.isabs

        if (
            os_path_abspath is None or os_path_isabs is None or rPath is None
        ):  # Interpreter shutdown
            return filename, filename

        isabs = os_path_isabs(filename)

        abs_path = _NormPath(filename, os_path_abspath, isabs)
        real_path = _NormPath(filename, rPath, isabs)

        # cache it for fast access later
        NORM_PATHS_CONTAINER[filename] = abs_path, real_path
        return abs_path, real_path


def _get_relative_filename_abs_path(filename, normpath, os_path_exists=os_path_exists):
    # If we have a relative path and the file does not exist when made absolute, try to
    # resolve it based on the sys.path entries.
    for p in sys.path:
        r = normpath(os.path.join(p, filename))
        if os_path_exists(r):
            return r

    # We couldn't find the real file for the relative path. Resolve it as if it was in
    # a library (so that it's considered a library file and not a project file).
    r = normpath(os.path.join(_library_dir, filename))
    return r


def _NormPath(filename, normpath, isabs, os_path_exists=os_path_exists, join=join):
    if filename.startswith("<"):
        # Not really a file, rather a synthetic name like <string> or <ipython-...>;
        # shouldn't be normalized.
        return filename

    r = normpath(filename)

    if not isabs:
        if not os_path_exists(r):
            r = _get_relative_filename_abs_path(filename, normpath)

    ind = r.find(".zip")
    if ind == -1:
        ind = r.find(".egg")
    if ind != -1:
        ind += 4
        zip_path = r[:ind]
        inner_path = r[ind:]
        if inner_path.startswith("!"):
            # Note (fabioz): although I can replicate this by creating a file ending as
            # .zip! or .egg!, I don't really know what's the real-world case for this
            # (still kept as it was added by @jetbrains, but it should probably be reviewed
            # later on).
            # Note 2: it goes hand-in-hand with 'exists'.
            inner_path = inner_path[1:]
            zip_path = zip_path + "!"

        if inner_path.startswith("/") or inner_path.startswith("\\"):
            inner_path = inner_path[1:]
        if inner_path:
            r = join(normcase(zip_path), inner_path)
            return r

    r = normcase(r)
    return r


_ZIP_SEARCH_CACHE: dict = {}
_NOT_FOUND_SENTINEL = object()


def exists(file):
    if os_path_exists(file):
        return True

    if not os.path.isabs(file):
        file = _get_relative_filename_abs_path(file, os.path.abspath)
        if os_path_exists(file):
            return True

    ind = file.find(".zip")
    if ind == -1:
        ind = file.find(".egg")

    if ind != -1:
        ind += 4
        zip_path = file[:ind]
        inner_path = file[ind:]
        if inner_path.startswith("!"):
            # Note (fabioz): although I can replicate this by creating a file ending as
            # .zip! or .egg!, I don't really know what's the real-world case for this
            # (still kept as it was added by @jetbrains, but it should probably be reviewed
            # later on).
            # Note 2: it goes hand-in-hand with '_NormPath'.
            inner_path = inner_path[1:]
            zip_path = zip_path + "!"

        zip_file_obj = _ZIP_SEARCH_CACHE.get(zip_path, _NOT_FOUND_SENTINEL)
        if zip_file_obj is None:
            return False
        elif zip_file_obj is _NOT_FOUND_SENTINEL:
            try:
                import zipfile

                zip_file_obj = zipfile.ZipFile(zip_path, "r")
                _ZIP_SEARCH_CACHE[zip_path] = zip_file_obj
            except:
                _ZIP_SEARCH_CACHE[zip_path] = _NOT_FOUND_SENTINEL
                return False

        try:
            if inner_path.startswith("/") or inner_path.startswith("\\"):
                inner_path = inner_path[1:]

            _info = zip_file_obj.getinfo(inner_path.replace("\\", "/"))

            return join(zip_path, inner_path)
        except KeyError:
            return False
    return False


# Note: as these functions may be rebound, users should always import
# pydevd_file_utils and then use:
#
# pydevd_file_utils.norm_file_to_client
# pydevd_file_utils.norm_file_to_server
#
# instead of importing any of those names to a given scope.


def _path_to_expected_str(filename):
    if isinstance(filename, bytes):
        filename = filename.decode(file_system_encoding)

    return filename


def _original_file_to_client(filename, cache={}):
    try:
        return cache[filename]
    except KeyError:
        translated = _path_to_expected_str(get_path_with_real_case(_AbsFile(filename)))
        cache[filename] = (translated, False)
    return cache[filename]


_original_file_to_server = _NormFile

norm_file_to_client = _original_file_to_client
norm_file_to_server = _original_file_to_server


def _fix_path(path, sep):
    if path.endswith("/") or path.endswith("\\"):
        path = path[:-1]

    if sep != "/":
        path = path.replace("/", sep)
    return path


_last_client_server_paths_set = []

_source_reference_to_frame_id = {}
_source_reference_to_server_filename: dict = {}
_line_cache_source_reference_to_server_filename = {}
_client_filename_in_utf8_to_source_reference: dict = {}
_next_source_reference = partial(next, itertools.count(1))


def get_client_filename_source_reference(client_filename):
    return _client_filename_in_utf8_to_source_reference.get(client_filename, 0)


def get_server_filename_from_source_reference(source_reference):
    return _source_reference_to_server_filename.get(source_reference, "")


def create_source_reference_for_linecache(server_filename):
    source_reference = _next_source_reference()
    _line_cache_source_reference_to_server_filename[source_reference] = server_filename
    return source_reference


def get_source_reference_filename_from_linecache(source_reference):
    return _line_cache_source_reference_to_server_filename.get(source_reference)


def create_source_reference_for_frame_id(frame_id):
    source_reference = _next_source_reference()
    _source_reference_to_frame_id[source_reference] = frame_id
    return source_reference


def get_frame_id_from_source_reference(source_reference):
    return _source_reference_to_frame_id.get(source_reference)


def setup_client_server_paths(paths):
    """paths is the same format as PATHS_FROM_IDE_TO_PYTHON"""

    global norm_file_to_client
    global norm_file_to_server
    global _last_client_server_paths_set
    global _next_source_reference

    _last_client_server_paths_set = paths[:]

    _source_reference_to_server_filename.clear()
    _client_filename_in_utf8_to_source_reference.clear()
    _next_source_reference = partial(next, itertools.count(1))

    # Work on the client and server slashes.
    python_sep = "\\" if IS_WINDOWS else "/"
    eclipse_sep = "\\" if _ide_os == "WINDOWS" else "/"

    norm_filename_to_server_container = {}
    norm_filename_to_client_container = {}
    initial_paths = list(paths)
    paths_from_IDE_to_python = initial_paths[:]

    # Apply normcase to the existing paths to follow the os preferences.

    for i, (path0, path1) in enumerate(paths_from_IDE_to_python[:]):
        path0 = _fix_path(path0, eclipse_sep)
        path1 = _fix_path(path1, python_sep)
        initial_paths[i] = (path0, path1)

        paths_from_IDE_to_python[i] = (_normcase_from_client(path0), normcase(path1))

    if not paths_from_IDE_to_python:
        # no translation step needed (just inline the calls)
        norm_file_to_client = _original_file_to_client
        norm_file_to_server = _original_file_to_server
        return

    # only setup translation functions if absolutely needed!
    def _norm_file_to_server(filename, cache=norm_filename_to_server_container):
        # Eclipse will send the passed filename to be translated to the python process
        # So, this would be 'NormFileFromEclipseToPython'
        try:
            return cache[filename]
        except KeyError:
            if eclipse_sep != python_sep:
                # Make sure that the separators are what we expect from the IDE.
                filename = filename.replace(python_sep, eclipse_sep)

            # used to translate a path from the client to the debug server
            translated = filename
            translated_normalized = _normcase_from_client(filename)
            for eclipse_prefix, server_prefix in paths_from_IDE_to_python:
                if translated_normalized.startswith(eclipse_prefix):
                    found_translation = True
                    if DEBUG_CLIENT_SERVER_TRANSLATION:
                        log.critical(
                            "robot debugger: replacing to server: %s", filename
                        )
                    translated = server_prefix + filename[len(eclipse_prefix) :]
                    if DEBUG_CLIENT_SERVER_TRANSLATION:
                        log.critical("robot debugger: sent to server: %s", translated)
                    break
            else:
                found_translation = False

            # Note that when going to the server, we do the replace first and only later do the norm file.
            if eclipse_sep != python_sep:
                translated = translated.replace(eclipse_sep, python_sep)

            if found_translation:
                translated = _NormFile(translated)
            else:
                if not os_path_exists(translated):
                    if not translated.startswith("<"):
                        # This is a configuration error, so, write it always so
                        # that the user can fix it.
                        log.error_once(
                            'robot debugger: unable to find translation for: "%s" in [%s] (please revise your path mappings).\n',
                            filename,
                            ", ".join(
                                ['"%s"' % (x[0],) for x in paths_from_IDE_to_python]
                            ),
                        )
                else:
                    # It's possible that we had some round trip (say, we sent /usr/lib and received
                    # it back, so, having no translation is ok too).
                    translated = _NormFile(translated)

            cache[filename] = translated
            return translated

    def _norm_file_to_client(filename, cache=norm_filename_to_client_container):
        # The result of this method will be passed to eclipse
        # So, this would be 'NormFileFromPythonToEclipse'
        try:
            return cache[filename]
        except KeyError:
            # used to translate a path from the debug server to the client
            translated = _NormFile(filename)

            # After getting the real path, let's get it with the path with
            # the real case and then obtain a new normalized copy, just in case
            # the path is different now.
            translated_proper_case = get_path_with_real_case(translated)
            translated = _NormFile(translated_proper_case)

            path_mapping_applied = False

            if IS_WINDOWS:
                if translated.lower() != translated_proper_case.lower():
                    translated_proper_case = translated
                    if DEBUG_CLIENT_SERVER_TRANSLATION:
                        log.critical(
                            "robot debugger: _NormFile changed path (from: %s to %s)",
                            translated_proper_case,
                            translated,
                        )

            for i, (eclipse_prefix, python_prefix) in enumerate(
                paths_from_IDE_to_python
            ):
                if translated.startswith(python_prefix):
                    if DEBUG_CLIENT_SERVER_TRANSLATION:
                        log.critical(
                            "robot debugger: replacing to client: %s", translated
                        )

                    # Note: use the non-normalized version.
                    eclipse_prefix = initial_paths[i][0]
                    translated = (
                        eclipse_prefix + translated_proper_case[len(python_prefix) :]
                    )
                    if DEBUG_CLIENT_SERVER_TRANSLATION:
                        log.critical("robot debugger: sent to client: %s", translated)
                    path_mapping_applied = True
                    break
            else:
                if DEBUG_CLIENT_SERVER_TRANSLATION:
                    log.critical(
                        "robot debugger: to client: unable to find matching prefix for: %s in %s",
                        translated,
                        [x[1] for x in paths_from_IDE_to_python],
                    )
                    translated = translated_proper_case

            if eclipse_sep != python_sep:
                translated = translated.replace(python_sep, eclipse_sep)

            translated = _path_to_expected_str(translated)

            # The resulting path is not in the python process, so, we cannot do a _NormFile here,
            # only at the beginning of this method.
            cache[filename] = (translated, path_mapping_applied)

            if translated not in _client_filename_in_utf8_to_source_reference:
                if path_mapping_applied:
                    source_reference = 0
                else:
                    source_reference = _next_source_reference()
                _client_filename_in_utf8_to_source_reference[
                    translated
                ] = source_reference
                _source_reference_to_server_filename[source_reference] = filename

            return (translated, path_mapping_applied)

    norm_file_to_server = _norm_file_to_server
    norm_file_to_client = _norm_file_to_client


setup_client_server_paths(PATHS_FROM_IDE_TO_PYTHON)


# For given file f returns tuple of its absolute path, real path and base name
def get_abs_path_real_path_and_base_from_file(
    filename, NORM_PATHS_AND_BASE_CONTAINER=NORM_PATHS_AND_BASE_CONTAINER
):
    try:
        return NORM_PATHS_AND_BASE_CONTAINER[filename]
    except:
        f = filename
        if not f:
            # i.e.: it's possible that the user compiled code with an empty string (consider
            # it as <string> in this case).
            f = "<string>"
        if _NormPaths is None:  # Interpreter shutdown
            i = max(f.rfind("/"), f.rfind("\\"))
            return (f, f, f[i + 1 :])

        if f is not None:
            if f.endswith(".pyc"):
                f = f[:-1]
            elif f.endswith("$py.class"):
                f = f[: -len("$py.class")] + ".py"

        abs_path, real_path = _NormPaths(f)

        try:
            base = basename(real_path)
        except AttributeError:
            # Error during shutdown.
            i = max(f.rfind("/"), f.rfind("\\"))
            base = f[i + 1 :]
        ret = abs_path, real_path, base
        NORM_PATHS_AND_BASE_CONTAINER[filename] = ret
        return ret


def get_abs_path_real_path_and_base_from_frame(frame):
    try:
        return NORM_PATHS_AND_BASE_CONTAINER[frame.f_code.co_filename]
    except:
        # This one is just internal (so, does not need any kind of client-server translation)
        f = frame.f_code.co_filename
        if f is not None and f.startswith(("build/bdist.", "build\\bdist.")):
            # files from eggs in Python 2.7 have paths like build/bdist.linux-x86_64/egg/<path-inside-egg>
            f = frame.f_globals["__file__"]

        if get_abs_path_real_path_and_base_from_file is None:
            # Interpreter shutdown
            if not f:
                # i.e.: it's possible that the user compiled code with an empty string (consider
                # it as <string> in this case).
                f = "<string>"
            i = max(f.rfind("/"), f.rfind("\\"))
            return f, f, f[i + 1 :]

        ret = get_abs_path_real_path_and_base_from_file(f)
        # Also cache based on the frame.f_code.co_filename (if we had it inside build/bdist it can make a difference).
        NORM_PATHS_AND_BASE_CONTAINER[frame.f_code.co_filename] = ret
        return ret


def get_fullname(mod_name):
    import pkgutil

    try:
        loader = pkgutil.get_loader(mod_name)
    except:
        return None
    if loader is not None:
        for attr in ("get_filename", "_get_filename"):
            meth = getattr(loader, attr, None)
            if meth is not None:
                return meth(mod_name)
    return None


def get_package_dir(mod_name):
    for path in sys.path:
        mod_path = join(path, mod_name.replace(".", "/"))
        if os.path.isdir(mod_path):
            return mod_path
    return None



"""
Comments checkers
"""
import re

from codecs import BOM_UTF8, BOM_UTF16_BE, BOM_UTF16_LE, BOM_UTF32_BE, BOM_UTF32_LE

from robot.api import Token
from robot.utils import FileReader

from robocop.checkers import RawFileChecker, VisitorChecker
from robocop.rules import Rule, RuleSeverity, RuleParam
from robocop.utils import ROBOT_VERSION
from robocop.exceptions import ConfigGeneralError

def regex(value):
    converted = rf'{value}'
    try:
        return re.compile(converted)
    except re.error as regex_err:
        raise ValueError(f'Regex error: {regex_err}')

rules = {
    "0701": Rule(
        RuleParam(
            name="markers",
            default="todo,fixme",
            converter=str,
            desc="List of case-insensitive markers that violate the rule in comments.",
        ),
        rule_id="0701",
        name="todo-in-comment",
        msg="Found a marker '{{ marker }}' in the comments",
        severity=RuleSeverity.WARNING,
        docs="""
        Report occurrences of the configured, case-insensitive marker in the comments.
        By default, it reports TODO and FIXME markers.

        Example::
        
            # TODO: Refactor this code
            # fixme
        
        Configuration example::

            robocop --configure "todo-in-comment:markers:todo,Remove me,Fix this!"

        """,
    ),
    "0702": Rule(
        RuleParam(
            name="block",
            default="^###",
            converter=regex,
            desc="Block comment regex pattern.",
        ),
        rule_id="0702",
        name="missing-space-after-comment",
        msg="Missing blank space after comment character",
        severity=RuleSeverity.WARNING,
        docs="""
        Make sure to have one blank space after '#' comment character.
        Configured regex for block comment should take into account the first character is `#`.

        Example::
        
            #bad
            # good
            ### good block

        Configuration example::

            robocop --configure missing-space-after-comment:block:^#[*]+

            Allows commenting like:

                #*****
                #
                # Important topics here!
                #
                #*****
                or
                #* Headers *#

        """,
    ),
    "0703": Rule(
        rule_id="0703",
        name="invalid-comment",
        msg="Invalid comment. '#' needs to be first character in the cell. "
        "For block comments you can use '*** Comments ***' section",
        severity=RuleSeverity.ERROR,
        version="<4.0",
        docs="""
        In Robot Framework 3.2.2 comments that started from second character in the cell were not recognized as 
        comments.
        
        Example::
        
            # good
             # bad
              # third cell so it's good
        
        """,
    ),
    "0704": Rule(
        rule_id="0704",
        name="ignored-data",
        msg="Ignored data found in file",
        severity=RuleSeverity.WARNING,
        docs="""
        All lines before first test data section 
        (`ref <https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-data-sections>`_) 
        are ignored. It's recommended to add `*** Comments ***` section header for lines that should be ignored.
        
        Missing section header::
    
            Resource   file.resource  # it looks like *** Settings *** but section header is missing - line is ignored
            
            *** Keywords ***
            Keyword Name
               No Operation
        
        Comment lines that should be inside `*** Comments ***`::
            
            Deprecated Test
                Keyword
                Keyword 2
            
            *** Test Cases ***
    
        """,
    ),
    "0705": Rule(
        rule_id="0705",
        name="bom-encoding-in-file",
        msg="This file contains BOM (Byte Order Mark) encoding not supported by Robot Framework",
        severity=RuleSeverity.WARNING,
        docs="""
        Some code editors can save Robot file using BOM encoding. Ensure that file is saved in UTF-8 encoding.
        """,
    ),
}


class CommentChecker(VisitorChecker):
    """Checker for comments content. It detects invalid comments or leftovers like `todo` or `fixme` in the code."""

    reports = (
        "todo-in-comment",
        "missing-space-after-comment",
        "invalid-comment",
    )

    def __init__(self):
        self._markers = None
        self._block = None
        super().__init__()

    @property
    def markers(self):
        if not self._markers:
            self._markers = self.param("todo-in-comment", "markers").lower().split(",")
        return self._markers

    @property
    def block(self):
        if not self._block:
            self._block = self.param("missing-space-after-comment", "block")
        return self._block

    def visit_Comment(self, node):  # noqa
        self.find_comments(node)

    def visit_TestCase(self, node):  # noqa
        self.check_invalid_comments(node.name, node)
        self.generic_visit(node)

    visit_Keyword = visit_TestCase

    def visit_Statement(self, node):  # noqa
        self.find_comments(node)

    def find_comments(self, node):
        """
        Find comments in node and check them for validity.
        Line can have only one comment, but the comment can contain separators.
        If the comment have separator it will be recognized as COMMENT, SEPARATOR, COMMENT in AST.
        We need to merge such comments into one for validity checks.
        """
        for line in node.lines:
            first_comment = None
            merged_comment = ""
            prev_sep = ""
            for token in line:
                if token.type == Token.SEPARATOR:
                    prev_sep = token.value
                elif token.type == Token.COMMENT:
                    if first_comment:
                        merged_comment += prev_sep + token.value
                    else:
                        merged_comment = token.value
                        first_comment = token
            if first_comment:
                self.check_comment_content(first_comment, merged_comment)

    def check_invalid_comments(self, name, node):
        if ROBOT_VERSION.major != 3:
            return
        if name and name.lstrip().startswith("#"):
            hash_pos = name.find("#")
            self.report("invalid-comment", node=node, col=node.col_offset + hash_pos + 1)

    def check_comment_content(self, token, content):
        low_content = content.lower()
        for violation in [marker for marker in self.markers if marker in low_content]:
            index = low_content.find(violation)
            self.report(
                "todo-in-comment",
                marker=content[index:index+len(violation)],
                lineno=token.lineno,
                col=token.col_offset + 1 + index,
            )
        if content.startswith("#") and not self.is_block_comment(content):
            if not content.startswith("# "):
                self.report(
                    "missing-space-after-comment",
                    lineno=token.lineno,
                    col=token.col_offset + 1,
                )

    def is_block_comment(self, comment):
        return comment == "#" or self.block.match(comment) is not None


class IgnoredDataChecker(RawFileChecker):
    """Checker for ignored data."""

    reports = (
        "ignored-data",
        "bom-encoding-in-file",
    )
    BOM = [BOM_UTF32_BE, BOM_UTF32_LE, BOM_UTF8, BOM_UTF16_LE, BOM_UTF16_BE]

    def __init__(self):
        self.is_bom = False
        super().__init__()

    def parse_file(self):
        self.is_bom = False
        if self.lines is not None:
            for lineno, line in enumerate(self.lines, start=1):
                if self.check_line(line, lineno):
                    break
        else:
            self.detect_bom(self.source)
            with FileReader(self.source) as file_reader:
                for lineno, line in enumerate(file_reader.readlines(), start=1):
                    if self.check_line(line, lineno):
                        break

    def check_line(self, line, lineno):
        if line.startswith("***"):
            return True
        if not line.startswith("# robocop:"):
            if lineno == 1 and self.is_bom:
                # if it's BOM encoded file, first line can be ignored
                return "***" in line
            self.report("ignored-data", lineno=lineno, col=1)
            return True
        return False

    def detect_bom(self, source):
        with open(source, "rb") as raw_file:
            first_four = raw_file.read(4)
            self.is_bom = any(first_four.startswith(bom_marker) for bom_marker in IgnoredDataChecker.BOM)
            if self.is_bom:
                self.report("bom-encoding-in-file", lineno=1, col=1)



from robotframework_ls.impl.protocols import ICompletionContext, ILocalizationInfo
from typing import Optional
from robotframework_ls.impl.text_utilities import normalize_robot_name


class _Requisites(object):
    def __init__(self, section, matcher, replace_from_col, replace_to_col, selection):
        self.section = section
        self.matcher = matcher
        self.replace_from_col = replace_from_col
        self.replace_to_col = replace_to_col
        self.selection = selection


def get_requisites(completion_context: ICompletionContext) -> Optional[_Requisites]:
    section_node = completion_context.get_ast_current_section()
    if section_node is None:
        return None

    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robotframework_ls.impl.section_completions import get_section_constant

    section = get_section_constant(completion_context, section_node)
    if section is None:
        return None

    selection = completion_context.sel  #: :type selection: DocumentSelection
    line_to_col = selection.line_to_column
    if line_to_col.endswith("  "):
        return None
    replace_to_col = selection.col
    if section.names_in_brackets:
        for i, c in enumerate(line_to_col):
            if c.isspace():
                continue
            elif c == "[":
                line_to_col = line_to_col[i + 1 :]
                replace_from_col = i
                break
            else:
                return None
        else:
            return None

        matcher = RobotStringMatcher(line_to_col)

    else:
        # i.e.: Needs to be the first char
        matcher = RobotStringMatcher(line_to_col)
        replace_from_col = 0

    return _Requisites(section, matcher, replace_from_col, replace_to_col, selection)


def complete(completion_context: ICompletionContext):
    from robocorp_ls_core.lsp import (
        TextEdit,
        Range,
        Position,
        CompletionItem,
        CompletionItemKind,
    )
    from robotframework_ls.impl.robot_version import robot_version_supports_language

    requisites = get_requisites(completion_context)
    if requisites is None:
        return []

    section = requisites.section
    matcher = requisites.matcher
    replace_from_col = requisites.replace_from_col
    selection = requisites.selection
    replace_to_col = requisites.replace_to_col

    line = selection.current_line
    sel_ends_with_close = line[selection.col :].startswith("]")

    ret = []
    if robot_version_supports_language():
        from robot.api import Language

        locinfo: ILocalizationInfo = completion_context.get_ast_localization_info()
        current_section_name = normalize_robot_name(
            completion_context.get_current_section_name()
        )

        def _translated_words():
            lang: Language
            for lang in locinfo.iter_languages_on_write():
                markers = section.markers_for_lang(lang)
                for marker in markers:
                    if normalize_robot_name(marker) == current_section_name:
                        yield from iter(section.names_for_lang(lang))
                        return

            # If it didn't return (is this possible?), provide all.
            for lang in locinfo.iter_languages_on_write():
                yield from iter(section.names_for_lang(lang))

        words = tuple(_translated_words())

    else:
        words = section.get_names_in_section_pre_rf_5_1()

    for word in sorted(words):
        if matcher.accepts(word):
            col_delta = 0
            if section.names_in_brackets:
                label = f"[{word}]"
                replacement = label
                if sel_ends_with_close:
                    col_delta = 1

            else:
                label = word
                replacement = word

            text_edit = TextEdit(
                Range(
                    start=Position(selection.line, replace_from_col),
                    end=Position(selection.line, replace_to_col + col_delta),
                ),
                replacement,
            )
            # text_edit = None
            ret.append(
                CompletionItem(
                    label, kind=CompletionItemKind.Keyword, text_edit=text_edit
                ).to_dict()
            )

    return ret



from robot.api.parsing import Token
from robot.utils.normalizing import normalize_whitespace

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.transformers import Transformer


class NormalizeSettingName(Transformer):
    """
    Normalize setting name.
    Ensure that setting names are title case without leading or trailing whitespace. For example from:

    ```robotframework
    *** Settings ***
    library    library.py
    test template    Template
    FORCE taGS    tag1

    *** Keywords ***
    Keyword
        [arguments]    ${arg}
        [ DOCUMENTATION]   Setup Keyword
    ```

    To:

    ```robotframework
    *** Settings ***
    Library    library.py
    Test Template    Template
    Force Tags    tag1

    *** Keywords ***
    Keyword
        [Arguments]    ${arg}
        [Documentation]   Setup Keyword
    ```
    """

    def __init__(self):
        super().__init__()  # workaround for our dynamically imported classes with args from cli/config

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    @skip_if_disabled
    def visit_Statement(self, node):  # noqa
        if node.type not in Token.SETTING_TOKENS:
            return node
        name = node.data_tokens[0].value
        if name.startswith("["):
            name = f"[{self.normalize_name(name[1:-1])}]"
        else:
            name = self.normalize_name(name)
        node.data_tokens[0].value = name
        return node

    @staticmethod
    def normalize_name(name):
        return normalize_whitespace(name).strip().title()



import os
from contextlib import contextmanager
from typing import List, Any, Dict, Set, Tuple, Optional
from robocorp_ls_core.protocols import TypedDict

from robocorp_ls_core.basic import isinstance_name
from robotframework_ls.impl.protocols import ICompletionContext, IKeywordFound, INode
from robotframework_ls.impl.text_utilities import (
    normalize_robot_name as get_internal_name,
)


class _UserKeywordType(TypedDict):
    type: str
    name: str
    internal_name: str
    kind: str
    doc: str
    args: List[str]
    body: List[Dict]


class _KeywordRecursionStack:
    _stack: Set[str]

    def __init__(self) -> None:
        self._stack = set()

    def __contains__(self, keyword_name: str) -> bool:
        return get_internal_name(keyword_name) in self._stack

    @contextmanager
    def scoped(self, keyword_name):
        self._stack.add(get_internal_name(keyword_name))
        yield
        self._stack.remove(get_internal_name(keyword_name))


class _UserKeywordCollector:
    _user_keywords: List[_UserKeywordType]
    _name_collection: Set[str]

    def __init__(self) -> None:
        self._user_keywords = []
        self._name_collection = set()

    @property
    def keywords(self) -> List[_UserKeywordType]:
        return self._user_keywords

    def __contains__(self, keyword_name: str) -> bool:
        return get_internal_name(keyword_name) in self._name_collection

    def append(self, keyword: _UserKeywordType) -> None:
        if "name" in keyword:
            self._user_keywords.append(keyword)
            self._name_collection.add(get_internal_name(keyword["name"]))


def _compute_suite_name(completion_context: ICompletionContext) -> str:
    suite_name = os.path.splitext(os.path.basename(completion_context.doc.uri))[0]
    suite_name = suite_name.title()
    return suite_name


def build_flow_explorer_model(completion_contexts: List[ICompletionContext]) -> dict:
    from robotframework_ls.impl import ast_utils

    suites: list = []
    recursion_stack: _KeywordRecursionStack = _KeywordRecursionStack()

    for completion_context in completion_contexts:
        ast = completion_context.get_ast()

        # Uncomment to print ast.
        # ast_utils.print_ast(ast)

        if ast:
            user_keywords_collector = _UserKeywordCollector()
            suite_name = _compute_suite_name(completion_context)
            tasks: list = []
            keywords: list = user_keywords_collector.keywords
            suite = {
                "type": "suite",
                "name": suite_name,
                "source": completion_context.doc.uri,
                "tasks": tasks,
                "keywords": keywords,
                "setup": None,
                "teardown": None,
            }
            suites.append(suite)
            for test in ast_utils.iter_tests(ast):
                test_name = f"{test.node.name} ({suite_name.lower()})"
                test_body: list = []
                test_info = {
                    "type": "task",
                    "name": test_name,
                    "internal_name": get_internal_name(test_name),
                    "doc": "",
                    "setup": None,
                    "teardown": None,
                    "body": test_body,
                }
                tasks.append(test_info)
                for node_info in ast_utils.iter_all_nodes(test.node, recursive=False):
                    with recursion_stack.scoped(test_name):
                        _build_hierarchy(
                            completion_context=completion_context,
                            curr_stack=node_info.stack,
                            curr_ast=node_info.node,
                            suite_name=suite_name,
                            parent_body=test_body,
                            memo={},
                            recursion_stack=recursion_stack,
                            user_keywords_collector=user_keywords_collector,
                            parent_node=test_info,
                        )
            for user_keyword in ast_utils.iter_keywords(ast):
                user_keyword_name = f"{user_keyword.node.name} ({suite_name.lower()})"
                user_keyword_body: list = []
                user_keyword_info = {
                    "type": "user-keyword",
                    "kind": "implemented",
                    "name": user_keyword_name,
                    "internal_name": get_internal_name(user_keyword_name),
                    "doc": "",
                    "body": user_keyword_body,
                }

                # Keywords var will be populated when building hierarchy if importing statements
                # Checking to see if it already exists before appending
                if get_internal_name(user_keyword_name) not in user_keywords_collector:
                    keywords.append(user_keyword_info)
                    for node_info in ast_utils.iter_all_nodes(
                        user_keyword.node, recursive=False
                    ):
                        with recursion_stack.scoped(user_keyword_name):
                            _build_hierarchy(
                                completion_context=completion_context,
                                curr_stack=node_info.stack,
                                curr_ast=node_info.node,
                                suite_name=suite_name,
                                parent_body=user_keyword_body,
                                memo={},
                                recursion_stack=recursion_stack,
                                user_keywords_collector=user_keywords_collector,
                            )

    if not suites:
        return {}

    # Reorder to the expected structure where we must specify the root suite.
    root_suite = suites[0]
    if len(suites) > 1:
        root_suite["suites"] = suites[1:]

    return root_suite


def _build_hierarchy(
    completion_context: ICompletionContext,
    curr_stack: Tuple[INode, ...],
    curr_ast: Any,
    suite_name: str,
    parent_body: List[Any],
    memo: dict,
    recursion_stack: _KeywordRecursionStack,
    user_keywords_collector: _UserKeywordCollector,
    parent_node: Optional[Dict] = None,
):
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl import ast_utils_keyword_usage
    from robotframework_ls.impl.find_definition import find_keyword_definition
    from robotframework_ls.impl.protocols import TokenInfo

    if ast_utils.is_keyword_usage_node(curr_ast):
        keyword_usage_handler = ast_utils_keyword_usage.obtain_keyword_usage_handler(
            curr_stack, curr_ast, recursive=True
        )
        if keyword_usage_handler is not None:
            for keyword_usage in keyword_usage_handler.iter_keyword_usages_from_node():
                keyword_body: list = []
                keyword_usage_node: Any = keyword_usage.node

                if isinstance_name(keyword_usage_node, "KeywordCall"):
                    keyword = {
                        "type": "keyword",
                        "kind": "simple",
                        "assign": keyword_usage_node.assign,
                        "args": keyword_usage_node.args,
                        "body": keyword_body,
                        "doc": "",
                    }
                    parent_body.append(keyword)

                    # Now, we need to follow the keyword and build its own structure
                    token_info = TokenInfo(
                        keyword_usage.stack, keyword_usage.node, keyword_usage.token
                    )
                    definitions = find_keyword_definition(
                        completion_context.create_copy_with_selection(
                            keyword_usage.token.lineno - 1,
                            keyword_usage.token.col_offset,
                        ),
                        token_info,
                    )

                    # Fallback name if we don't know where it's defined.
                    keyword["name"] = f"{keyword_usage_node.keyword}"
                    keyword["internal_name"] = get_internal_name(keyword["name"])
                    if definitions:
                        # Use the first one
                        definition = next(iter(definitions))
                        keyword_found: IKeywordFound = definition.keyword_found
                        if keyword_found.library_name:
                            keyword[
                                "name"
                            ] = f"{keyword_usage_node.keyword} ({keyword_found.library_name.lower()})"
                        elif keyword_found.resource_name:
                            keyword[
                                "name"
                            ] = f"{keyword_usage_node.keyword} ({keyword_found.resource_name.lower()})"
                        keyword["internal_name"] = get_internal_name(keyword["name"])

                        # If it was found in a library we don't recurse anymore.
                        keyword_ast = keyword_found.keyword_ast
                        if keyword_ast is None:
                            continue
                        definition_completion_context = keyword_found.completion_context
                        if definition_completion_context is None:
                            continue
                        # If found in recursion stack we don't recurse anymore.
                        if keyword["name"] in recursion_stack:
                            keyword["kind"] = "recursion-leaf"
                            continue
                        suite_name = _compute_suite_name(definition_completion_context)
                        # Ok, it isn't a library keyword (as we have its AST). Keep recursing.
                        for node_info in ast_utils.iter_all_nodes(
                            keyword_ast, recursive=False
                        ):
                            with recursion_stack.scoped(keyword["name"]):
                                _build_hierarchy(
                                    completion_context=definition_completion_context,
                                    curr_stack=node_info.stack,
                                    curr_ast=node_info.node,
                                    suite_name=suite_name,
                                    parent_body=keyword_body,
                                    memo=memo,
                                    recursion_stack=recursion_stack,
                                    user_keywords_collector=user_keywords_collector,
                                )
                        # If the current keyword has body, the it is a User Keyword
                        if (
                            len(keyword_body) > 0
                            and get_internal_name(keyword["name"])
                            not in user_keywords_collector
                        ):
                            user_keyword: _UserKeywordType = {
                                "type": "user-keyword",
                                "kind": "implemented",
                                "body": keyword["body"],
                                "name": keyword["name"],
                                "internal_name": keyword["internal_name"],
                                "doc": keyword["doc"],
                                "args": keyword["args"],
                            }
                            user_keywords_collector.append(user_keyword)
                elif isinstance_name(keyword_usage_node, "Teardown") and parent_node:
                    parent_node["teardown"] = {
                        "type": "keyword",
                        "subtype": "KEYWORD",
                        "args": keyword_usage_node.args,
                        "name": keyword_usage_node.name,
                    }
                elif isinstance_name(keyword_usage_node, "Setup") and parent_node:
                    parent_node["setup"] = {
                        "type": "keyword",
                        "subtype": "KEYWORD",
                        "args": keyword_usage_node.args,
                        "name": keyword_usage_node.name,
                    }
    elif isinstance_name(curr_ast, "If"):
        if_body: list = []
        if_info: Dict[str, Any] = {"type": "if", "body": if_body}
        parent_body.append(if_info)

        condition = " ".join(
            str(tok) for tok in ast_utils.iter_argument_tokens(curr_ast.header)
        )
        if_branch_body: list = []
        if_branch_info: Dict[str, Any] = {
            "type": "if-branch",
            "condition": condition,
            "body": if_branch_body,
        }
        if_body.append(if_branch_info)
        for body_ast in curr_ast.body:
            _build_hierarchy(
                completion_context=completion_context,
                curr_stack=curr_stack + (curr_ast,),
                curr_ast=body_ast,
                suite_name=suite_name,
                parent_body=if_branch_body,
                memo=memo,
                recursion_stack=recursion_stack,
                user_keywords_collector=user_keywords_collector,
            )

        orelse = curr_ast.orelse

        def explore_elseifs(elseifbranch):
            if elseifbranch and isinstance_name(elseifbranch.header, "ElseIfHeader"):
                condition = " ".join(
                    str(tok)
                    for tok in ast_utils.iter_argument_tokens(elseifbranch.header)
                )
                else_if_body: list = []
                else_if_info: Dict[str, Any] = {
                    "type": "else-if-branch",
                    "condition": condition,
                    "body": else_if_body,
                }

                if_body.append(else_if_info)
                for body_ast in elseifbranch.body:
                    _build_hierarchy(
                        completion_context=completion_context,
                        curr_stack=curr_stack + (elseifbranch,),
                        curr_ast=body_ast,
                        suite_name=suite_name,
                        parent_body=else_if_body,
                        memo=memo,
                        recursion_stack=recursion_stack,
                        user_keywords_collector=user_keywords_collector,
                    )
                elseifbranch = elseifbranch.orelse
                if elseifbranch and isinstance_name(
                    elseifbranch.header, "ElseIfHeader"
                ):
                    explore_elseifs(elseifbranch)

        explore_elseifs(orelse)

        # To finish, handle the orelse.
        orelse = orelse.orelse if orelse and orelse.orelse else orelse
        if orelse:
            orelse_body: list = []
            orelse_info: Dict[str, Any] = {
                "type": "else-branch",
                "body": orelse_body,
            }
            if_body.append(orelse_info)
            for body_ast in orelse.body:
                _build_hierarchy(
                    completion_context=completion_context,
                    curr_stack=curr_stack + (orelse,),
                    curr_ast=body_ast,
                    suite_name=suite_name,
                    parent_body=orelse_body,
                    memo=memo,
                    recursion_stack=recursion_stack,
                    user_keywords_collector=user_keywords_collector,
                )
    elif isinstance_name(curr_ast, "For"):
        for_body: list = []
        for_info: Dict[str, Any] = {
            "type": "for",
            "kind": curr_ast.flavor,
            "values": list(curr_ast.values),
            "variables": list(curr_ast.variables),
            "body": for_body,
        }
        parent_body.append(for_info)
        for body_ast in curr_ast.body:
            _build_hierarchy(
                completion_context=completion_context,
                curr_stack=curr_stack + (curr_ast,),
                curr_ast=body_ast,
                suite_name=suite_name,
                parent_body=for_body,
                memo=memo,
                recursion_stack=recursion_stack,
                user_keywords_collector=user_keywords_collector,
            )
    elif isinstance_name(curr_ast, "While"):
        condition = " ".join(
            str(tok) for tok in ast_utils.iter_argument_tokens(curr_ast.header)
        )
        while_body: list = []
        while_info: Dict[str, Any] = {
            "type": "while",
            "condition": condition,
            "body": while_body,
        }
        parent_body.append(while_info)
        for body_ast in curr_ast.body:
            _build_hierarchy(
                completion_context=completion_context,
                curr_stack=curr_stack + (curr_ast,),
                curr_ast=body_ast,
                suite_name=suite_name,
                parent_body=while_body,
                memo=memo,
                recursion_stack=recursion_stack,
                user_keywords_collector=user_keywords_collector,
            )
    elif isinstance_name(curr_ast, "Try"):
        try_body: list = []
        try_info: Dict[str, Any] = {
            "type": "try",
            "body": try_body,
        }
        parent_body.append(try_info)
        try_branch_body: list = []
        try_branch_info: Dict[str, Any] = {
            "type": "try-branch",
            "body": try_branch_body,
        }
        try_body.append(try_branch_info)

        for body_ast in curr_ast.body:
            _build_hierarchy(
                completion_context=completion_context,
                curr_stack=curr_stack + (curr_ast,),
                curr_ast=body_ast,
                suite_name=suite_name,
                parent_body=try_branch_body,
                memo=memo,
                recursion_stack=recursion_stack,
                user_keywords_collector=user_keywords_collector,
            )

        def explore_try(ast):
            if ast:
                next_type = None
                next_patterns = None
                if isinstance_name(ast.header, "ExceptHeader"):
                    next_type = "except-branch"
                    next_patterns = ast.patterns
                elif isinstance_name(ast.header, "FinallyHeader"):
                    next_type = "finally-branch"
                elif isinstance_name(ast.header, "ElseHeader"):
                    next_type = "else-branch"
                if not next_type:
                    return
                next_branch_body: list = []
                next_branch_info: Dict[str, Any] = {
                    "type": next_type,
                    "body": next_branch_body,
                }
                if next_patterns:
                    next_branch_info["patterns"] = next_patterns
                for body_ast in ast.body:
                    _build_hierarchy(
                        completion_context=completion_context,
                        curr_stack=curr_stack + (ast,),
                        curr_ast=body_ast,
                        suite_name=suite_name,
                        parent_body=next_branch_body,
                        memo=memo,
                        recursion_stack=recursion_stack,
                        user_keywords_collector=user_keywords_collector,
                    )
                try_body.append(next_branch_info)
                if ast.next:
                    explore_try(ast.next)

        explore_try(curr_ast.next)
    elif isinstance_name(curr_ast, "Break"):
        break_info: Dict[str, Any] = {
            "type": "break",
        }
        parent_body.append(break_info)
    elif isinstance_name(curr_ast, "Continue"):
        continue_info: Dict[str, Any] = {
            "type": "continue",
        }
        parent_body.append(continue_info)
    elif isinstance_name(curr_ast, "ReturnStatement"):
        return_info: Dict[str, Any] = {
            "type": "return",
        }
        parent_body.append(return_info)



from robocorp_ls_core.robotframework_log import get_logger
from typing import Tuple

log = get_logger(__name__)


def get_robot_version() -> str:
    try:
        import robot  # noqa

        v = str(robot.get_version())
    except:
        log.exception("Unable to get robot version.")
        v = "unknown"
    return v


def _get_robot_naked_version():
    try:
        import robot  # noqa

        v = str(robot.get_version(True))
    except:
        log.exception("Unable to get robot version.")
        v = "unknown"
    return v


_found_major_version = None


def get_robot_major_version() -> int:
    global _found_major_version
    if _found_major_version is not None:
        return _found_major_version

    robot_version = _get_robot_naked_version()

    major_version = 4
    try:
        if "." in robot_version:
            major_version = int(robot_version.split(".")[0])
            _found_major_version = major_version
    except:
        log.exception("Unable to get robot major version.")

    return major_version


_found_major_minor_version = None


def get_robot_major_minor_version() -> Tuple[int, int]:
    global _found_major_minor_version
    if _found_major_minor_version is not None:
        return _found_major_minor_version

    robot_version = _get_robot_naked_version()

    major_minor_version = (4, 0)
    try:
        if "." in robot_version:
            split = robot_version.split(".")
            major_version = int(split[0])
            minor_version = int(split[1])
            major_minor_version = _found_major_minor_version = (
                major_version,
                minor_version,
            )
    except:
        log.exception("Unable to get robot major/minor version.")

    return major_minor_version


def robot_version_supports_language():
    return get_robot_major_minor_version() >= (5, 1)



__version__ = "1.7.3"
version_info = [int(x) for x in __version__.split(".")]

import os.path
import sys

__file__ = os.path.abspath(__file__)
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]


def import_robocorp_ls_core():
    """
    Helper function to make sure that robocorp_ls_core is imported properly
    (either in dev or in release mode).
    """

    try:
        import robocorp_ls_core
    except ImportError:
        log_contents = []
        use_folder = None
        try:
            src_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            log_contents.append("Source folder: %s" % (src_folder,))
            src_core_folder = os.path.abspath(
                os.path.join(src_folder, "..", "..", "robocorp-python-ls-core", "src")
            )

            if os.path.isdir(src_core_folder):
                log_contents.append("Dev mode detected. Found: %s" % (src_core_folder,))
                use_folder = src_core_folder

            else:
                vendored_folder = os.path.join(
                    src_folder, "robotframework_ls", "vendored"
                )
                log_contents.append(
                    "Using vendored mode. Found: %s" % (vendored_folder,)
                )
                use_folder = vendored_folder
                assert os.path.isdir(
                    use_folder
                ), "Expected: %s to exist and be a directory." % (use_folder,)

            sys.path.append(use_folder)
            import robocorp_ls_core
        except:
            try:
                if use_folder:
                    log_contents.append(
                        "%s contents:\n%s" % (use_folder, os.listdir(use_folder))
                    )
            except:
                log_contents.append("Error in os.listdir('%s')." % (use_folder,))
            raise ImportError(
                "Error importing robocorp_ls_core. Log: %s" % "\n".join(log_contents)
            )


def _import_helper(import_callback, project_name):
    try:
        import_callback()
    except ImportError:
        log_contents = []
        try:
            src_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            log_contents.append("Source folder: %s" % (src_folder,))

            vendored_folder = os.path.join(src_folder, "robotframework_ls", "vendored")
            log_contents.append(f"Using vendored mode. Found: {vendored_folder}")
            use_folder = vendored_folder
            assert os.path.isdir(
                use_folder
            ), f"Expected: {use_folder} to exist and be a directory."

            if use_folder not in sys.path:
                sys.path.append(use_folder)
            try:
                import_callback()
            except ImportError:
                log_contents.append(
                    f"Unable to import {project_name} from vendored folder."
                )

            check_path = [src_folder, "..", "..", project_name, "src"]
            src_core_folder = os.path.abspath(os.path.join(*check_path))
            if not os.path.exists(src_core_folder):
                check_path = [src_folder, "..", "..", "..", project_name, "src"]
                src_core_folder = os.path.abspath(os.path.join(*check_path))

            log_contents.append(f"Dev mode detected. Using: {src_core_folder}")
            use_folder = src_core_folder

            if use_folder not in sys.path:
                sys.path.append(use_folder)
            import_callback()
        except:
            try:
                if use_folder:
                    log_contents.append(
                        "%s contents:\n%s" % (use_folder, os.listdir(use_folder))
                    )
            except:
                log_contents.append("Error in os.listdir('%s')." % (use_folder,))
            raise ImportError(
                "Error importing robocorp_ls_core. Log: %s" % "\n".join(log_contents)
            )


def import_rf_interactive():
    """
    Helper function to make sure that robotframework_interactive is imported properly
    (either in dev or in release mode).
    """

    def import_callback():
        import robotframework_interactive

    _import_helper(import_callback, "robotframework-interactive")


def import_robot_out_stream():
    """
    Helper function to make sure that robot_out_stream is imported properly
    (either in dev or in release mode).

    Note:
    In dev mode we expect the:
    "robotframework-output-stream"
    project to be checked out right next to the
    "robotframework-lsp"
    project.
    """

    def import_callback():
        import robot_out_stream

    _import_helper(import_callback, "robotframework-output-stream")



import os
import sys
from difflib import unified_diff
from typing import Optional, Pattern, Tuple

try:
    import rich_click as click
except ImportError:  # Fails on vendored-in LSP plugin
    import click

    escape = None

from robot.api import get_model
from robot.errors import DataError

from robotidy.config import Config
from robotidy.disablers import RegisterDisablers
from robotidy.utils import ModelWriter, StatementLinesCollector, decorate_diff_with_color, escape_rich_markup

try:
    from robotidy.rich_console import console
except ImportError:  # Fails on vendored-in LSP plugin
    console = None


class Robotidy:
    def __init__(self, config: Config):
        self.config = config

    def transform_files(self):
        changed_files = 0
        disabler_finder = RegisterDisablers(self.config.formatting.start_line, self.config.formatting.end_line)
        for source in self.config.sources:
            try:
                stdin = False
                if str(source) == "-":
                    stdin = True
                    if self.config.verbose:
                        click.echo("Loading file from stdin")
                    source = self.load_from_stdin()
                elif self.config.verbose:
                    click.echo(f"Transforming {source} file")
                model = get_model(source)
                disabler_finder.visit(model)
                if disabler_finder.file_disabled:
                    continue
                diff, old_model, new_model = self.transform(model, disabler_finder.disablers)
                if diff:
                    changed_files += 1
                self.output_diff(model.source, old_model, new_model)
                if stdin:
                    self.print_to_stdout(new_model)
                elif diff:
                    self.save_model(model.source, model)
            except DataError:
                click.echo(
                    f"Failed to decode {source}. Default supported encoding by Robot Framework is UTF-8. Skipping file"
                )
                pass
        if not self.config.check or not changed_files:
            return 0
        return 1

    def transform(self, model, disablers):
        old_model = StatementLinesCollector(model)
        for transformer in self.config.transformers:
            setattr(transformer, "disablers", disablers)  # set dynamically to allow using external transformers
            transformer.visit(model)
        new_model = StatementLinesCollector(model)
        return new_model != old_model, old_model, new_model

    @staticmethod
    def load_from_stdin() -> str:
        return sys.stdin.read()

    def print_to_stdout(self, collected_lines):
        if not self.config.show_diff:
            click.echo(collected_lines.text)

    def save_model(self, source, model):
        if self.config.overwrite:
            output = self.config.output or model.source
            ModelWriter(output=output, newline=self.get_line_ending(source)).write(model)

    def get_line_ending(self, path: str):
        if self.config.formatting.line_sep == "auto":
            with open(path) as f:
                f.readline()
                if f.newlines is None:
                    return os.linesep
                if isinstance(f.newlines, str):
                    return f.newlines
                else:
                    return f.newlines[0]
        return self.config.formatting.line_sep

    def output_diff(
        self,
        path: str,
        old_model: StatementLinesCollector,
        new_model: StatementLinesCollector,
    ):
        if not self.config.show_diff:
            return
        old = [l + "\n" for l in old_model.text.splitlines()]
        new = [l + "\n" for l in new_model.text.splitlines()]
        lines = list(unified_diff(old, new, fromfile=f"{path}\tbefore", tofile=f"{path}\tafter"))
        if not lines:
            return
        if self.config.color:
            output = decorate_diff_with_color(lines)
        else:
            output = escape_rich_markup(lines)
        for line in output:
            console.print(line, end="", highlight=False, soft_wrap=True)



from typing import List, Dict, Any, Iterator, Optional
import typing

from robocorp_ls_core.lsp import (
    CommandTypedDict,
    ICustomDiagnosticDataTypedDict,
    ICustomDiagnosticDataUndefinedKeywordTypedDict,
    WorkspaceEditTypedDict,
    CompletionItemTypedDict,
    TextEditTypedDict,
    WorkspaceEditParamsTypedDict,
    ICustomDiagnosticDataUndefinedResourceTypedDict,
    RangeTypedDict,
    ShowDocumentParamsTypedDict,
    ICustomDiagnosticDataUndefinedLibraryTypedDict,
    ICustomDiagnosticDataUndefinedVarImportTypedDict,
    ICustomDiagnosticDataUnexpectedArgumentTypedDict,
)
from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IKeywordFound,
    IResourceImportNode,
    IRobotDocument,
)
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.basic import isinstance_name
import os
from pathlib import Path
from robotframework_ls.impl.robot_generated_lsp_constants import (
    OPTION_ROBOT_QUICK_FIX_KEYWORD_TEMPLATE,
)

log = get_logger(__name__)


def _add_import_code_action(
    completion_context: ICompletionContext,
) -> Iterator[CommandTypedDict]:
    from robotframework_ls.impl.collect_keywords import (
        collect_keyword_name_to_keyword_found,
    )
    from robotframework_ls.impl import auto_import_completions

    keyword_name_to_keyword_found: Dict[
        str, List[IKeywordFound]
    ] = collect_keyword_name_to_keyword_found(completion_context)
    auto_imports_found: List[
        CompletionItemTypedDict
    ] = auto_import_completions.complete(
        completion_context,
        keyword_name_to_keyword_found,
        use_for_quick_fix=True,
    )

    for auto_import in auto_imports_found:
        label = auto_import["label"]
        if label.endswith("*"):
            label = label[:-1]

        lst: List[TextEditTypedDict] = []

        text_edit = auto_import["textEdit"]
        if text_edit:
            lst.append(text_edit)

        additional = auto_import["additionalTextEdits"]
        if additional:
            lst.extend(additional)

        changes = {completion_context.doc.uri: lst}
        edit: WorkspaceEditTypedDict = {"changes": changes}
        title = f"Import {label}"
        edit_params: WorkspaceEditParamsTypedDict = {"edit": edit, "label": title}
        command: CommandTypedDict = {
            "title": title,
            "command": "robot.applyCodeAction",
            "arguments": [{"apply_edit": edit_params}],
        }

        yield command


def _create_keyword_in_current_file_text_edit(
    completion_context: ICompletionContext,
    keyword_template: str,
) -> TextEditTypedDict:
    from robotframework_ls.impl import ast_utils

    current_section: Any = completion_context.get_ast_current_section()
    if ast_utils.is_keyword_section(current_section):
        # Add it before the current keyword
        use_line = None
        for node in current_section.body:
            if isinstance_name(node, "Keyword"):
                node_lineno = node.lineno - 1

                if node_lineno <= completion_context.sel.line:
                    use_line = node_lineno
                else:
                    break

        if use_line is not None:
            return {
                "range": {
                    "start": {"line": use_line, "character": 0},
                    "end": {"line": use_line, "character": 0},
                },
                "newText": keyword_template,
            }

    keyword_section = ast_utils.find_keyword_section(completion_context.get_ast())
    if keyword_section is None:
        # We need to create the keyword section too
        current_section = completion_context.get_ast_current_section()
        if current_section is None:
            use_line = 0
        else:
            use_line = current_section.lineno - 1

        return {
            "range": {
                "start": {"line": use_line, "character": 0},
                "end": {"line": use_line, "character": 0},
            },
            "newText": f"*** Keywords ***\n{keyword_template}",
        }

    else:
        # We add the keyword after the end of the existing keyword section
        use_line = keyword_section.end_lineno
        if completion_context.doc.get_line(use_line - 1).strip():
            keyword_template = "\n\n" + keyword_template

        elif completion_context.doc.get_line(use_line - 2).strip():
            keyword_template = "\n" + keyword_template

        return {
            "range": {
                "start": {"line": use_line, "character": 0},
                "end": {"line": use_line, "character": 0},
            },
            "newText": keyword_template,
        }


def _create_keyword_in_current_file_code_action(
    completion_context: ICompletionContext, keyword_template: str, keyword_name: str
) -> Iterator[CommandTypedDict]:

    text_edit = _create_keyword_in_current_file_text_edit(
        completion_context, keyword_template
    )
    lst: List[TextEditTypedDict] = [text_edit]

    changes = {completion_context.doc.uri: lst}
    edit: WorkspaceEditTypedDict = {"changes": changes}
    title = f"Create Keyword: {keyword_name} (in current file)"
    edit_params: WorkspaceEditParamsTypedDict = {"edit": edit, "label": title}
    command: CommandTypedDict = {
        "title": title,
        "command": "robot.applyCodeAction",
        "arguments": [{"apply_edit": edit_params}],
    }

    _add_show_document_at_command(command, completion_context.doc.uri, text_edit)
    yield command


def _undefined_resource_code_action(
    completion_context: ICompletionContext,
    undefined_resource_data: ICustomDiagnosticDataUndefinedResourceTypedDict,
) -> Iterator[CommandTypedDict]:
    from robocorp_ls_core.lsp import CreateFileTypedDict
    from robocorp_ls_core import uris

    name = undefined_resource_data["resolved_name"]
    if not name:
        name = undefined_resource_data["name"]
        if not name:
            return

    if "$" in name or "{" in name or "}" in name:
        return

    path = Path(os.path.join(os.path.dirname(completion_context.doc.path), name))
    doc_uri = uris.from_fs_path(str(path))
    create_doc_change: CreateFileTypedDict = {
        "kind": "create",
        "uri": doc_uri,
    }
    edit: WorkspaceEditTypedDict = {"documentChanges": [create_doc_change]}
    title: str = f"Create {path.name} (at {path.parent})"
    edit_params: WorkspaceEditParamsTypedDict = {"edit": edit, "label": title}

    command: CommandTypedDict = {
        "title": title,
        "command": "robot.applyCodeAction",
        "arguments": [{"apply_edit": edit_params}],
    }

    _add_show_document_at_command(command, doc_uri)

    yield command


def _undefined_keyword_code_action(
    completion_context: ICompletionContext,
    undefined_keyword_data: ICustomDiagnosticDataUndefinedKeywordTypedDict,
) -> Iterator[CommandTypedDict]:
    from robotframework_ls.robot_config import get_arguments_separator

    keyword_template = "$keyword_name$keyword_arguments\n    $cursor\n\n"
    config = completion_context.config
    if config is not None:
        keyword_template = config.get_setting(
            OPTION_ROBOT_QUICK_FIX_KEYWORD_TEMPLATE, str, keyword_template
        )

    # Make it less likely that we'll have conflicts for our variables.
    keyword_template = keyword_template.replace("$cursor", "$__LSP_CURSOR_LOCATION__$")
    keyword_template = keyword_template.replace(
        "$keyword_name", "$__LSP_KEYWORD_NAME_LOCATION__$"
    )
    keyword_template = keyword_template.replace(
        "$keyword_arguments", "$__LSP_KEYWORD_ARGUMENTS_LOCATION__$"
    )

    # --- Update the arguments in the template.

    arguments: List[str] = []
    keyword_usage_info = completion_context.get_current_keyword_usage_info()
    if keyword_usage_info is not None:
        for token in keyword_usage_info.node.tokens:
            if token.type == token.ARGUMENT:
                i = token.value.find("=")
                if i > 0:
                    name = token.value[:i]
                else:
                    name = token.value
                if not name:
                    name = "arg"
                arguments.append(f"${{{name}}}")

    separator = get_arguments_separator(completion_context)
    args_str = ""
    if arguments:
        args_str += "\n    [Arguments]"
        for arg in arguments:
            args_str += separator
            args_str += arg

    keyword_template = keyword_template.replace(
        "$__LSP_KEYWORD_ARGUMENTS_LOCATION__$", args_str
    )

    # --- Update the keyword name in the template.

    # We'd like to have a cursor here, but alas, this isn't possible...
    # See: https://github.com/microsoft/language-server-protocol/issues/592
    # See: https://github.com/microsoft/language-server-protocol/issues/724
    keyword_name = undefined_keyword_data["name"]

    dots_found = keyword_name.count(".")
    if dots_found >= 2:
        # Must check for use cases... Do nothing for now.
        return

    if dots_found == 1:
        # Something as:
        # my_resource.Keyword or
        # my_python_module.Keyword
        #
        # in this case we need to create a keyword "Keyword" in "my_resource".
        # If my_module is imported, create it in that module, otherwise,
        # if it exists but we haven't imported it, we need to import it.
        # If it doesn't exist we need to create it first.
        splitted = keyword_name.split(".")
        resource_or_import_or_alias_name, keyword_name = splitted
        keyword_template = keyword_template.replace(
            "$__LSP_KEYWORD_NAME_LOCATION__$", keyword_name
        )
        yield from _deal_with_resource_or_import_or_alias_name(
            completion_context,
            resource_or_import_or_alias_name,
            keyword_template,
            keyword_name,
        )
        return

    keyword_template = keyword_template.replace(
        "$__LSP_KEYWORD_NAME_LOCATION__$", keyword_name
    )

    yield from _add_import_code_action(completion_context)
    yield from _create_keyword_in_current_file_code_action(
        completion_context, keyword_template, keyword_name
    )


def _matches_resource_import(
    resource_import: IResourceImportNode,
    name: str,
):
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    name = normalize_robot_name(name)

    for token in resource_import.tokens:
        if token.type == token.NAME:
            import_name = normalize_robot_name(token.value)

            if import_name == name:
                return True

            # ./my_resource.robot -> my_resource.robot
            import_name = os.path.basename(import_name)
            if import_name == name:
                return True

            # Handle something as my_resource.robot
            import_name = os.path.splitext(import_name)[0]
            if import_name == name:
                return True

    return False


def _create_keyword_in_another_file_code_action(
    completion_context: ICompletionContext, keyword_template: str, keyword_name: str
) -> Iterator[CommandTypedDict]:

    text_edit = _create_keyword_in_current_file_text_edit(
        completion_context, keyword_template
    )
    lst: List[TextEditTypedDict] = [text_edit]

    changes = {completion_context.doc.uri: lst}
    edit: WorkspaceEditTypedDict = {"changes": changes}
    modname: str = os.path.basename(completion_context.doc.uri)
    title = f"Create Keyword: {keyword_name} (in {modname})"
    edit_params: WorkspaceEditParamsTypedDict = {"edit": edit, "label": title}

    command: CommandTypedDict = {
        "title": title,
        "command": "robot.applyCodeAction",
        "arguments": [
            {
                "apply_edit": edit_params,
            }
        ],
    }

    _add_show_document_at_command(command, completion_context.doc.uri, text_edit)

    yield command


def _add_show_document_at_command(
    command: CommandTypedDict,
    doc_uri: str,
    text_edit: Optional[TextEditTypedDict] = None,
):
    if text_edit:
        new_text = text_edit["newText"]
        cursor_i = new_text.find("$__LSP_CURSOR_LOCATION__$")
        if cursor_i == -1:
            endline = text_edit["range"]["end"]["line"]
            endchar = text_edit["range"]["end"]["character"]
        else:
            endline = text_edit["range"]["start"]["line"]
            endchar = text_edit["range"]["start"]["character"]
            # Find the actual cursor_i location (and remove it from the text)
            text_edit["newText"] = new_text.replace("$__LSP_CURSOR_LOCATION__$", "", 1)
            for line_i, text in enumerate(new_text.splitlines()):
                if "$__LSP_CURSOR_LOCATION__$" in text:
                    endline += line_i
                    endchar += text.find("$__LSP_CURSOR_LOCATION__$")
                    break

    else:
        endline = 0
        endchar = 0

    selection: RangeTypedDict = {
        "start": {"line": endline, "character": endchar},
        "end": {"line": endline, "character": endchar},
    }
    show_document: ShowDocumentParamsTypedDict = {
        "uri": doc_uri,
        "selection": selection,
        "takeFocus": True,
    }

    arguments = command["arguments"]
    if arguments:
        arguments[0]["show_document"] = show_document


def _deal_with_resource_or_import_or_alias_name(
    completion_context: ICompletionContext,
    resource_or_import_or_alias_name: str,
    keyword_template: str,
    keyword_name: str,
):
    for resource_import in completion_context.get_resource_imports():
        if _matches_resource_import(resource_import, resource_or_import_or_alias_name):
            doc = completion_context.get_resource_import_as_doc(resource_import)
            if doc is not None:
                new_completion_context = completion_context.create_copy(doc)
                yield from _create_keyword_in_another_file_code_action(
                    new_completion_context, keyword_template, keyword_name
                )


def _create_arguments_command(
    completion_context: ICompletionContext,
    use_line: int,
    use_col: int,
    arg_name: str,
    keyword_name: str,
    prefix: str = "",
    postfix: str = "",
) -> CommandTypedDict:
    from robotframework_ls.robot_config import get_arguments_separator

    separator = get_arguments_separator(completion_context)
    text_edit: TextEditTypedDict = {
        "range": {
            "start": {
                "line": use_line,
                "character": use_col,
            },
            "end": {
                "line": use_line,
                "character": use_col,
            },
        },
        "newText": f"{prefix}{separator}{arg_name}$__LSP_CURSOR_LOCATION__${postfix}",
    }

    changes = {completion_context.doc.uri: [text_edit]}
    edit: WorkspaceEditTypedDict = {"changes": changes}
    title = f"Add argument {arg_name} to {keyword_name}"
    edit_params: WorkspaceEditParamsTypedDict = {
        "edit": edit,
        "label": title,
    }
    command: CommandTypedDict = {
        "title": title,
        "command": "robot.applyCodeAction",
        "arguments": [{"apply_edit": edit_params}],
    }

    _add_show_document_at_command(command, completion_context.doc.uri, text_edit)
    return command


def _unexpected_argument_code_action(
    completion_context: ICompletionContext,
    unexpected_argument_data: ICustomDiagnosticDataUnexpectedArgumentTypedDict,
):
    from robocorp_ls_core import uris
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.string_matcher import RobotStringMatcher

    arg_name = unexpected_argument_data["arg_name"]
    if "=" not in arg_name:
        return

    arg_name = arg_name.split("=")[0]
    if not arg_name:
        return

    arg_name = "${%s}" % (arg_name,)

    keyword_name = unexpected_argument_data["keyword_name"]
    path = unexpected_argument_data["path"]
    doc = completion_context.workspace.get_document(
        uris.from_fs_path(path), accept_from_file=True
    )
    if not doc:
        return

    robotdoc = typing.cast(IRobotDocument, doc)
    completion_context = completion_context.create_copy(robotdoc)
    ast = robotdoc.get_ast()
    robot_string_matcher = RobotStringMatcher(keyword_name)
    matched_keyword_node = None
    for keyword in ast_utils.iter_keywords(ast):
        keyword_name = keyword.node.name
        if robot_string_matcher.is_keyword_name_match(keyword_name):
            matched_keyword_node = keyword.node
            break
    else:
        return

    for stmt in matched_keyword_node.body:
        if isinstance_name(stmt, "Arguments"):
            # Ok, found existing arguments
            for token in reversed(stmt.tokens):
                if token.type in (token.ARGUMENT, token.ARGUMENTS):
                    last_arg = token
                    use_line = last_arg.lineno - 1
                    use_col = last_arg.end_col_offset

                    yield _create_arguments_command(
                        completion_context,
                        use_line,
                        use_col,
                        arg_name,
                        keyword_name,
                    )
                    return

    # If we got here theres no [Arguments] section. So, create it along
    # with the arguments.
    header = matched_keyword_node.header
    prefix = f"    [Arguments]"

    yield _create_arguments_command(
        completion_context,
        use_line=header.end_lineno,
        use_col=0,
        arg_name=arg_name,
        keyword_name=keyword_name,
        prefix=prefix,
        postfix="\n",
    )


def code_action(
    completion_context: ICompletionContext,
    found_data: List[ICustomDiagnosticDataTypedDict],
) -> List[CommandTypedDict]:
    """
    Note: the completion context selection should be at the range end position.
    """

    ret: List[CommandTypedDict] = []
    for data in found_data:
        if data["kind"] == "undefined_keyword":
            undefined_keyword_data = typing.cast(
                ICustomDiagnosticDataUndefinedKeywordTypedDict, data
            )
            ret.extend(
                _undefined_keyword_code_action(
                    completion_context, undefined_keyword_data
                )
            )

        elif data["kind"] == "unexpected_argument":
            unexpected_argument_data = typing.cast(
                ICustomDiagnosticDataUnexpectedArgumentTypedDict, data
            )
            ret.extend(
                _unexpected_argument_code_action(
                    completion_context, unexpected_argument_data
                )
            )

        elif data["kind"] == "undefined_resource":
            undefined_resource_data = typing.cast(
                ICustomDiagnosticDataUndefinedResourceTypedDict, data
            )
            ret.extend(
                _undefined_resource_code_action(
                    completion_context, undefined_resource_data
                )
            )

        elif data["kind"] == "undefined_library":
            undefined_library_data = typing.cast(
                ICustomDiagnosticDataUndefinedLibraryTypedDict, data
            )
            ret.extend(
                _undefined_resource_code_action(
                    completion_context, undefined_library_data
                )
            )

        elif data["kind"] == "undefined_var_import":
            undefined_var_import_data = typing.cast(
                ICustomDiagnosticDataUndefinedVarImportTypedDict, data
            )
            ret.extend(
                _undefined_resource_code_action(
                    completion_context, undefined_var_import_data
                )
            )

    for r in ret:
        if r["command"] == "robot.applyCodeAction":
            arguments = r["arguments"]
            if arguments:
                arg = arguments[0]
                lint_uris = arg.get("lint_uris")
                if lint_uris is None:
                    lint_uris = []
                    arg["lint_uris"] = lint_uris
                lint_uris.append(completion_context.doc.uri)

    return ret



from typing import Optional, Iterator, Tuple

from robotframework_ls.impl.protocols import KeywordUsageInfo, IRobotToken
from robotframework_ls.impl.keywords_in_args import KEYWORD_NAME_TO_KEYWORD_INDEX
from robotframework_ls.impl.keywords_in_args import KEYWORD_NAME_TO_CONDITION_INDEX

TOK_TYPE_NONE = 0
TOK_TYPE_KEYWORD = 1
TOK_TYPE_EXPRESSION = 2
TOK_TYPE_CONTROL = 3
TOK_TYPE_IGNORE = 4


def _tok_type_as_str(tok_type) -> str:
    if tok_type == TOK_TYPE_NONE:
        return "<none>"
    if tok_type == TOK_TYPE_EXPRESSION:
        return "<expression>"
    if tok_type == TOK_TYPE_KEYWORD:
        return "<keyword>"
    if tok_type == TOK_TYPE_CONTROL:
        return "<control>"
    if tok_type == TOK_TYPE_IGNORE:
        return "<ignore>"
    raise AssertionError(f"Unexpected: {tok_type}")


class _ConsiderArgsAsKeywordNames:
    NONE = TOK_TYPE_NONE
    KEYWORD = TOK_TYPE_KEYWORD
    EXPRESSION = TOK_TYPE_EXPRESSION
    CONTROL = TOK_TYPE_CONTROL
    IGNORE = TOK_TYPE_IGNORE

    def __init__(
        self,
        node,
        normalized_keyword_name,
        consider_keyword_at_index,
        consider_condition_at_index,
    ):
        self._normalized_keyword_name = normalized_keyword_name
        self._consider_keyword_at_index = consider_keyword_at_index
        self._consider_condition_at_index = consider_condition_at_index
        self._current_arg = 0

        # Run Keyword If is special because it has 'ELSE IF' / 'ELSE'
        # which will then be be (cond, keyword) or just (keyword), so
        # we need to provide keyword usages as needed.
        if self._normalized_keyword_name == "runkeywordif":
            self.next_tok_type = self._next_tok_type_run_keyword_if
        elif self._normalized_keyword_name == "foreachinputworkitem":
            self.next_tok_type = self._next_tok_type_for_each_input_work_item
        elif self._normalized_keyword_name == "runkeywords":
            found = False
            for token in node.tokens:
                if "AND" == token.value:
                    found = True
                    break
            if found:
                self.next_tok_type = self._next_tok_type_run_keywords
            else:
                self.next_tok_type = self._consider_each_arg_as_keyword

        self._stack_kind = None
        self._stack = None
        self._started_match = False

    def next_tok_type(self, token) -> int:  # pylint: disable=method-hidden
        assert token.type == token.ARGUMENT
        self._current_arg += 1

        if self._current_arg == self._consider_condition_at_index:
            return self.EXPRESSION

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD

        return self.NONE

    def _next_tok_type_for_each_input_work_item(self, token):
        from robotframework_ls.impl.variable_resolve import find_split_index
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        assert token.type == token.ARGUMENT
        self._current_arg += 1

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD

        i = find_split_index(token.value)
        if i > 0:
            v = normalize_robot_name(token.value[:i])
            if v in ("itemslimit", "returnresults"):
                return self.IGNORE

        return self.NONE

    def _next_tok_type_run_keyword_if(self, token):
        assert token.type == token.ARGUMENT

        self._current_arg += 1

        if token.value == "ELSE IF":
            self._started_match = True
            self._stack = []
            self._stack_kind = token.value
            return self.CONTROL
        elif token.value == "ELSE":
            self._started_match = True
            self._stack = []
            self._stack_kind = token.value
            return self.CONTROL

        else:
            self._started_match = False
            if self._stack is not None:
                self._stack.append(token)

        if self._stack is not None:
            if self._stack_kind == "ELSE IF":
                if len(self._stack) == 1:
                    return self.EXPRESSION
                return self.KEYWORD if len(self._stack) == 2 else self.NONE

            if self._stack_kind == "ELSE":
                return self.KEYWORD if len(self._stack) == 1 else self.NONE

        if self._current_arg == self._consider_condition_at_index:
            return self.EXPRESSION

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD

        return self.NONE

    def _consider_each_arg_as_keyword(self, token):
        assert token.type == token.ARGUMENT
        return self.KEYWORD

    def _next_tok_type_run_keywords(self, token):
        assert token.type == token.ARGUMENT

        self._current_arg += 1

        if token.value == "AND":
            self._started_match = True
            self._stack = []
            self._stack_kind = token.value
            return self.CONTROL

        else:
            self._started_match = False
            if self._stack is not None:
                self._stack.append(token)

        if self._stack is not None:
            if self._stack_kind == "AND":
                return self.KEYWORD if len(self._stack) == 1 else self.NONE

        if self._current_arg == self._consider_keyword_at_index:
            return self.KEYWORD
        return self.NONE


def _create_root_keyword_usage_info(stack, node) -> Optional[KeywordUsageInfo]:
    """
    If this is a keyword usage node, return information on it, otherwise,
    returns None.

    :note: this goes hand-in-hand with get_keyword_name_token.
    """
    from robot.api import Token
    from robotframework_ls.impl.ast_utils import (
        CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET,
    )
    from robotframework_ls.impl.ast_utils import _strip_node_and_token_bdd_prefix

    if node.__class__.__name__ == "KeywordCall":
        token_type = Token.KEYWORD

    elif node.__class__.__name__ in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET:
        token_type = Token.NAME

    else:
        return None

    prefix, node, token = _strip_node_and_token_bdd_prefix(stack, node, token_type)
    if token is None:
        return None

    keyword_name = token.value
    if keyword_name.lower() == "none":
        return None
    return KeywordUsageInfo(tuple(stack), node, token, keyword_name, prefix=prefix)


def _build_keyword_usage(stack, node, current_tokens) -> Optional[KeywordUsageInfo]:
    from robotframework_ls.impl.ast_utils import copy_token_replacing

    # Note: just check for line/col because the token could be changed
    # (for instance, an EOL ' ' could be added to the token).
    if not current_tokens:
        return None

    keyword_at_index = 0
    keyword_token = current_tokens[keyword_at_index]

    keyword_token = copy_token_replacing(keyword_token, type=keyword_token.KEYWORD)
    new_tokens = [keyword_token]
    new_tokens.extend(current_tokens[keyword_at_index + 1 :])

    new_node = node.__class__(new_tokens)
    return KeywordUsageInfo(
        stack,
        new_node,
        keyword_token,
        keyword_token.value,
        True,
    )


def _iter_keyword_usage_info_uncached_from_args(
    stack, node, args_as_keywords_handler, token_line_col_to_type
) -> Iterator[KeywordUsageInfo]:
    # We may have multiple matches, so, we need to setup the appropriate book-keeping
    current_tokens = []

    iter_in = iter(node.tokens)

    for token in iter_in:
        if token.type == token.ARGUMENT:
            next_tok_type = args_as_keywords_handler.next_tok_type(token)
            token_line_col_to_type[(token.lineno, token.col_offset)] = next_tok_type
            if next_tok_type == args_as_keywords_handler.KEYWORD:
                current_tokens.append(token)
                break

    for token in iter_in:
        if token.type == token.ARGUMENT:
            next_tok_type = args_as_keywords_handler.next_tok_type(token)
            token_line_col_to_type[(token.lineno, token.col_offset)] = next_tok_type

            if next_tok_type in (
                args_as_keywords_handler.CONTROL,
                args_as_keywords_handler.EXPRESSION,
                args_as_keywords_handler.IGNORE,
            ):
                # Don't add IF/ELSE IF/AND nor the condition.
                continue

            if next_tok_type != args_as_keywords_handler.KEYWORD:
                # Argument was now added to current_tokens.
                current_tokens.append(token)
                continue

            if current_tokens:
                # Starting a new one (build for the previous).
                usage_info = _build_keyword_usage(
                    stack,
                    node,
                    current_tokens,
                )
                if usage_info is not None:
                    yield usage_info

            current_tokens = [token]

    else:
        # Do one last iteration at the end to deal with the last one.
        if current_tokens:
            usage_info = _build_keyword_usage(
                stack,
                node,
                current_tokens,
            )
            if usage_info is not None:
                yield usage_info


class _KeywordUsageHandler:
    """
    We have the following main use-cases when dealing with keyword usages (also
    known as keyword references):

    1. Obtain the usages (keyword call/arguments) for code-analysis.
    2. For each token in a keyword usage, know what it maps to (
       keyword name, expression, control, regular argument, ...)

    Also, it needs to be considered that a given keyword usage may have
    other usages within it, so, the _KeywordUsageHandler is an API to help
    make things more streamlined for each use-case.
    """

    NONE = TOK_TYPE_NONE
    KEYWORD = TOK_TYPE_KEYWORD
    EXPRESSION = TOK_TYPE_EXPRESSION
    CONTROL = TOK_TYPE_CONTROL
    IGNORE = TOK_TYPE_IGNORE

    def __init__(self, stack, node, recursive):
        self.node = node
        self.stack = stack
        self._recursive = recursive

        # We store as line/col the type info and not the actual token because we
        # may create dummy tokens along the way and in this case we're
        # interested in the positions.
        self._token_line_col_to_type = {}
        self._keyword_usages_from_node_cache = None

    def _ensure_cached(self):
        if self._keyword_usages_from_node_cache is None:
            self._keyword_usages_from_node_cache = tuple(
                self._iter_keyword_usages_from_node()
            )

    def iter_keyword_usages_from_node(self) -> Iterator[KeywordUsageInfo]:
        self._ensure_cached()
        yield from iter(self._keyword_usages_from_node_cache)

    def _iter_keyword_usages_from_node(self) -> Iterator[KeywordUsageInfo]:
        """
        Note: the iteration order is guaranteed and it's from the inside to
        the outside (because when matching tokens we want to match more
        specific ones before outer ones).
        """

        root_keyword_usage_info = _create_root_keyword_usage_info(self.stack, self.node)
        if root_keyword_usage_info is None:
            return

        # Ok, we have the root one, now, we need to recursively detect others.
        if self._recursive:
            yield from self._iter_keyword_usages_inside_keyword_usage(
                root_keyword_usage_info
            )

        yield root_keyword_usage_info

    def _iter_keyword_usages_inside_keyword_usage(
        self, root_keyword_usage_info: KeywordUsageInfo
    ) -> Iterator[KeywordUsageInfo]:
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        # Now, we have the root, determine if it can have other usages inside itself...
        normalized_keyword_name = normalize_robot_name(root_keyword_usage_info.name)
        consider_keyword_at_index = KEYWORD_NAME_TO_KEYWORD_INDEX.get(
            normalized_keyword_name
        )
        consider_condition_at_index = KEYWORD_NAME_TO_CONDITION_INDEX.get(
            normalized_keyword_name
        )
        if (
            consider_keyword_at_index is not None
            or consider_condition_at_index is not None
        ):
            args_as_keywords_handler = _ConsiderArgsAsKeywordNames(
                root_keyword_usage_info.node,
                normalized_keyword_name,
                consider_keyword_at_index,
                consider_condition_at_index,
            )

            for kw_usage in _iter_keyword_usage_info_uncached_from_args(
                self.stack,
                root_keyword_usage_info.node,
                args_as_keywords_handler,
                self._token_line_col_to_type,
            ):
                yield from self._iter_keyword_usages_inside_keyword_usage(kw_usage)
                yield kw_usage

    def get_token_type(self, tok: IRobotToken) -> int:
        """
        :return:
            TOK_TYPE_NONE = 0
            TOK_TYPE_KEYWORD = 1
            TOK_TYPE_EXPRESSION = 2
            TOK_TYPE_CONTROL = 3
            TOK_TYPE_IGNORE = 4
        """
        self._ensure_cached()
        return self._token_line_col_to_type.get(
            (tok.lineno, tok.col_offset), TOK_TYPE_NONE
        )

    def get_token_type_as_str(self, token: IRobotToken) -> str:
        return _tok_type_as_str(self.get_token_type(token))

    def iter_tokens_with_type(self) -> Iterator[Tuple[IRobotToken, int]]:
        self._ensure_cached()
        for tok in self.node.tokens:
            yield tok, self._token_line_col_to_type.get(
                (tok.lineno, tok.col_offset), TOK_TYPE_NONE
            )

    def get_keyword_usage_for_token_line_col(
        self, line, col
    ) -> Optional[KeywordUsageInfo]:
        self._ensure_cached()
        for kw_usage in self.iter_keyword_usages_from_node():
            for token in kw_usage.node.tokens:
                if token.lineno == line and token.col_offset == col:
                    return kw_usage
        return None


def obtain_keyword_usage_handler(
    stack, node, recursive=True
) -> Optional[_KeywordUsageHandler]:
    from robotframework_ls.impl.ast_utils import (
        CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET,
    )

    if (
        node.__class__.__name__ != "KeywordCall"
        and node.__class__.__name__
        not in CLASSES_WITH_ARGUMENTS_AS_KEYWORD_CALLS_AS_SET
    ):
        return None

    return _KeywordUsageHandler(stack, node, recursive=recursive)


def obtain_keyword_usage_for_token(stack, node, token) -> Optional[KeywordUsageInfo]:
    keyword_usage_handler = obtain_keyword_usage_handler(stack, node)
    if keyword_usage_handler is not None:
        keyword_usage = keyword_usage_handler.get_keyword_usage_for_token_line_col(
            token.lineno, token.col_offset
        )
        return keyword_usage
    return None



import os
from functools import lru_cache
from typing import Optional, Tuple, List, Iterator, Union
from robotframework_ls.impl.protocols import (
    IRobotVariableMatch,
    IRobotToken,
    ICompletionContext,
    AbstractVariablesCollector,
    IVariableFound,
)
from robocorp_ls_core.protocols import Sentinel
from robocorp_ls_core.robotframework_log import get_logger
import threading
import re

log = get_logger(__name__)


def is_number_var(normalized_variable_name):
    # see: robot.variables.finders.NumberFinder
    try:
        bases = {"0b": 2, "0o": 8, "0x": 16}
        if normalized_variable_name.startswith(tuple(bases)):
            int(normalized_variable_name[2:], bases[normalized_variable_name[:2]])
            return True
        int(normalized_variable_name)
        return True
    except:
        pass  # Let's try float...

    try:
        float(normalized_variable_name)
        return True
    except:
        pass

    return False


def is_python_eval_var(normalized_variable_name):
    return (
        len(normalized_variable_name) >= 2
        and normalized_variable_name[0] == "{"
        and normalized_variable_name[-1] == "}"
    )


_match_extended = re.compile(
    r"""
    (.+?)          # base name (group 1)
    ([^\s\w].+)    # extended part (group 2)
""",
    re.UNICODE | re.VERBOSE,
).match


def extract_var_name_from_extended_base_name(normalized_variable_name):
    m = _match_extended(normalized_variable_name)
    if m is None:
        return normalized_variable_name

    base_name, _extended = m.groups()
    return base_name


def extract_variable_base(text: str) -> str:
    """
    Converts something as: "${S_ome.VAR}[foo]" to "S_ome.VAR".
    """
    variable_match = robot_search_variable(text)
    if variable_match is not None:
        base = variable_match.base
        if base is not None:
            return base

    if len(text) >= 3:
        if (
            text.endswith("}")
            and text[1] == "{"
            and text.startswith(("$", "@", "&", "%"))
        ):
            return text[2:-1]
    return text


@lru_cache(maxsize=500)
def normalize_variable_name(text: str) -> str:
    """
    Converts something as: "${S_ome.VAR}[foo]" to "some.var".
    """
    base = extract_variable_base(text)

    return base.lower().replace("_", "").replace(" ", "")


@lru_cache(maxsize=200)
def robot_search_variable(text: str) -> Optional[IRobotVariableMatch]:
    """
    Provides the IRobotVariableMatch from a text such as "${S_ome.VAR}[foo]".
    """
    from robot.variables.search import search_variable  # type:ignore

    try:
        variable_match = search_variable(text, identifiers="$@&%", ignore_errors=True)
        return variable_match
    except:
        pass

    return None


def has_variable(text: str) -> bool:
    robot_match = robot_search_variable(text)
    if robot_match is None:
        return False
    return bool(robot_match.base)


def iter_robot_variable_matches(
    string: str,
) -> Iterator[Tuple[IRobotVariableMatch, int]]:
    """
    Provides the IRobotVariableMatch and the relative index for the match in the string.
    """
    # Based on robot.variables.search.VariableIterator
    remaining = string
    relative_index = 0
    while True:
        robot_match = robot_search_variable(remaining)
        if not robot_match:
            break
        yield robot_match, relative_index
        remaining = robot_match.after
        relative_index += len(robot_match.before) + len(robot_match.match)


def find_split_index(string: str) -> int:
    # Based on: robot.utils.escaping.split_from_equals
    # Based on: robot.utils.escaping._find_split_index

    eq_i = string.find("=")
    if eq_i == -1:
        return -1

    return _find_split_index(string, eq_i)


@lru_cache(maxsize=200)
def _find_split_index(string: str, eq_i: int) -> int:
    try:
        variables = tuple(iter_robot_variable_matches(string))
        if not variables and "\\" not in string:
            return eq_i

        relative_index = 0
        robot_match_len = 0
        for robot_match, relative_index in variables:
            before = robot_match.before
            string = robot_match.after
            try:
                return _find_split_index_from_part(before) + relative_index
            except ValueError:
                pass
            robot_match_len = robot_match.end - robot_match.start + len(before)

        return _find_split_index_from_part(string) + relative_index + robot_match_len
    except ValueError:
        return -1


def _find_split_index_from_part(string):
    index = 0
    while "=" in string[index:]:
        index += string[index:].index("=")
        if _not_escaping(string[:index]):
            return index
        index += 1
    raise ValueError()


def _not_escaping(name):
    backslashes = len(name) - len(name.rstrip("\\"))
    return backslashes % 2 == 0


class _VariablesCollector(AbstractVariablesCollector):
    def __init__(self):
        self.var_name_to_var_found = {}

    def accepts(self, variable_name: str) -> bool:
        return True

    def on_variable(self, variable_found: IVariableFound):
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        self.var_name_to_var_found[
            normalize_robot_name(variable_found.variable_name)
        ] = variable_found


class ResolveVariablesContext:
    _thread_local = threading.local()

    def __init__(self, completion_context: ICompletionContext):
        self.config = completion_context.config
        self.completion_context = completion_context

    @property
    def doc_path(self) -> str:
        return self.completion_context.doc.path

    def _resolve_environment_variable(self, var_name, value_if_not_found, log_info):
        ret = os.environ.get(var_name, Sentinel.SENTINEL)
        if ret is Sentinel.SENTINEL:
            log.info(*log_info)
            return value_if_not_found
        return ret

    def _convert_robot_variable(self, var_name, value_if_not_found):
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        normalized = normalize_robot_name(var_name)

        if normalized == "curdir":
            return os.path.dirname(self.doc_path)

        completion_context = self.completion_context

        # Settings and argument files have higher priority (arguments override things
        # in the document).
        found = completion_context.get_settings_normalized_var_name_to_var_found().get(
            normalized
        )
        if found is not None:
            return found.variable_value

        found = completion_context.get_arguments_files_normalized_var_name_to_var_found().get(
            normalized
        )
        if found is not None:
            return found.variable_value

        found = completion_context.get_doc_normalized_var_name_to_var_found().get(
            normalized
        )

        if found is not None:
            return found.variable_value

        found = completion_context.get_builtins_normalized_var_name_to_var_found(
            True
        ).get(normalized)
        if found is not None:
            return found.variable_value

        # At this point we have to do a search in the imports to check whether
        # maybe the variable is defined in a dependency. Note that we can get
        # into a recursion here as we may need to resolve other variables
        # in order to get here.
        # It goes something like:
        # Resolve variable -> some variable
        # collect dependencies (which may need to resolve variables again).
        # In this case we have a thread-local variable to know whether this is the
        # case and if it is bail out.
        try:
            resolve_info = self._thread_local.resolve_info
        except:
            resolve_info = self._thread_local.resolve_info = set()

        if var_name in resolve_info:
            log.info("Unable to find robot variable: %s", var_name)
            return value_if_not_found

        else:
            try:
                resolve_info.add(var_name)

                from robotframework_ls.impl.variable_completions import (
                    collect_global_variables_from_document_dependencies,
                )

                collector = _VariablesCollector()

                collect_global_variables_from_document_dependencies(
                    completion_context, collector
                )

                found = collector.var_name_to_var_found.get(normalized)
                if found is not None:
                    return found.variable_value

                log.info("Unable to find robot variable: %s", var_name)
                return value_if_not_found
            finally:
                resolve_info.discard(var_name)

    def _convert_environment_variable(
        self, var_name, value_if_not_found
    ) -> Optional[str]:
        from robotframework_ls.impl.robot_lsp_constants import OPTION_ROBOT_PYTHON_ENV

        i = var_name.find("=")
        if i > 0:
            value_if_not_found = var_name[i + 1 :]
            var_name = var_name[:i]

        if self.config is None:
            value = self._resolve_environment_variable(
                var_name,
                value_if_not_found,
                (
                    "Config not available while trying to convert environment variable: %s",
                    var_name,
                ),
            )
        else:
            robot_env_vars = self.config.get_setting(OPTION_ROBOT_PYTHON_ENV, dict, {})
            value = robot_env_vars.get(var_name, Sentinel.SENTINEL)
            if value is Sentinel.SENTINEL:
                value = self._resolve_environment_variable(
                    var_name,
                    value_if_not_found,
                    ("Unable to find environment variable: %s", var_name),
                )

        if value is None:
            return None
        return str(value)

    def token_value_resolving_variables(self, token: IRobotToken) -> str:
        return self.token_value_and_unresolved_resolving_variables(token)[0]

    def token_value_and_unresolved_resolving_variables(
        self, token: Union[IRobotToken, str], count=0
    ) -> Tuple[str, Tuple[Tuple[IRobotToken, str], ...]]:
        """
        :return: The new value of the token (with resolved variables) and if
        some variable was unresolved, a tuple of tuples with the token of the
        unresolved variable as well as the related error message.

        i.e.:
            (
                'resolved/var',
                (
                    (unresolved_token, error_msg), (unresolved_token, error_msg)
                )
            )
        """
        from robotframework_ls.impl import ast_utils

        robot_token: IRobotToken
        if isinstance(token, str):
            robot_token = ast_utils.create_token(token)
        else:
            robot_token = token

        try:
            tokenized_vars = ast_utils.tokenize_variables(robot_token)
        except:
            return robot_token.value, ()  # Unable to tokenize

        unresolved: List[Tuple[IRobotToken, str]] = []

        parts: List[str] = []
        tok: IRobotToken
        value: str
        for tok in tokenized_vars:
            if tok.type == tok.NAME:
                parts.append(str(tok))

            elif tok.type == tok.VARIABLE:
                if count > 6:
                    unresolved.append(
                        (
                            tok,
                            f"\nRecursion detected when resolving variable: {tok.value}.",
                        )
                    )
                    return str(tok), tuple(unresolved)
                value = str(tok)

                resolved, new_value = self._resolve_tok_var(value)
                if new_value is None:
                    new_value = value
                if not resolved:
                    unresolved.append(
                        (
                            tok,
                            f"\nUnable to statically resolve variable: {tok.value}.\nPlease set the `{tok.value[2:-1]}` value in `robot.variables`.",
                        )
                    )
                else:
                    # Ok, it was resolved, but we need to check if we need
                    # to resolve new values from that value (and do it
                    # recursively if needed).
                    for _ in range(6):
                        if "{" in new_value:
                            (
                                v2,
                                unresolved_tokens,
                            ) = self.token_value_and_unresolved_resolving_variables(
                                new_value, count + 1
                            )
                            if unresolved_tokens:
                                if len(unresolved_tokens) == 1:
                                    t = next(iter(unresolved_tokens))[0]
                                    unresolved.append(
                                        (
                                            tok,
                                            f"\nUnable to statically resolve variable: {tok.value} because dependent variable: {t.value} was not resolved.",
                                        )
                                    )
                                else:
                                    lst = []
                                    for t, _error_msg in unresolved_tokens:
                                        lst.append(t.value)

                                    unresolved.append(
                                        (
                                            tok,
                                            f"\nUnable to statically resolve variable: {tok.value} because dependent variables: {', '.join(lst)} were not resolved.",
                                        )
                                    )
                                break

                            if v2 == new_value:
                                break
                            new_value = v2
                parts.append(new_value)

        joined_parts = "".join(parts)
        return joined_parts, tuple(unresolved)

    def _resolve_tok_var(self, value: str) -> Tuple[bool, str]:
        """
        :return: whether the variable was resolved and its value (same as input if unresolved).
        """
        convert_with = None
        if value.startswith("${") and value.endswith("}"):
            convert_with = self._convert_robot_variable

        elif value.startswith("%{") and value.endswith("}"):
            convert_with = self._convert_environment_variable

        if convert_with is None:
            log.info("Cannot resolve variable: %s", value)
            return False, value  # Leave unresolved.
        else:
            inner_value = value[2:-1]
            converted = convert_with(inner_value, None)
            if converted is None:
                log.info("Cannot resolve variable: %s", value)
                return False, value
            else:
                return True, converted



from robot.api.parsing import (
    Comment,
    ElseHeader,
    ElseIfHeader,
    EmptyLine,
    End,
    ForHeader,
    IfHeader,
    ModelVisitor,
    Token,
)

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.transformers import Transformer
from robotidy.utils import is_suite_templated, round_to_four


class AlignTemplatedTestCases(Transformer):
    """
    Align templated Test Cases to columns.

    Following code:

    ```robotframework
    *** Test Cases ***    baz    qux
    # some comment
    test1    hi    hello
    test2 long test name    asdfasdf    asdsdfgsdfg
    ```

    will be transformed to:

    ```robotframework
    *** Test Cases ***      baz         qux
    # some comment
    test1                   hi          hello
    test2 long test name    asdfasdf    asdsdfgsdfg
                            bar1        bar2
    ```

    If you don't want to align test case section that does not contain header names (in above example baz and quz are
    header names) then configure `only_with_headers` parameter:

    ```
    robotidy -c AlignSettingsSection:only_with_hedaers:True <src>
    ```

    For non-templated test cases use ``AlignTestCasesSection`` transformer.
    """

    ENABLED = False

    def __init__(self, only_with_headers: bool = False, min_width: int = None):
        super().__init__()
        self.only_with_headers = only_with_headers
        self.min_width = min_width
        self.widths = None
        self.test_name_len = 0
        self.name_line = 0
        self.indent = 0

    def visit_File(self, node):  # noqa
        if not is_suite_templated(node):
            return node
        return self.generic_visit(node)

    def visit_If(self, node):  # noqa
        self.indent += 1
        self.generic_visit(node)
        self.indent -= 1
        return node

    visit_Else = visit_ElseIf = visit_For = visit_If

    @skip_section_if_disabled
    def visit_TestCaseSection(self, node):  # noqa
        if len(node.header.data_tokens) == 1 and self.only_with_headers:
            return node
        counter = ColumnWidthCounter(self.disablers)
        counter.visit(node)
        self.widths = counter.widths
        return self.generic_visit(node)

    @skip_if_disabled
    def visit_Statement(self, statement):  # noqa
        if statement.type == Token.TESTCASE_NAME:
            self.test_name_len = len(statement.data_tokens[0].value) if statement.data_tokens else 0
            self.name_line = statement.lineno
        elif statement.type == Token.TESTCASE_HEADER:
            self.align_header(statement)
        elif not isinstance(
            statement,
            (Comment, EmptyLine, ForHeader, IfHeader, ElseHeader, ElseIfHeader, End),
        ):
            self.align_statement(statement)
        return statement

    def align_header(self, statement):
        tokens = []
        # *** Test Cases ***            baz                            qux
        # *** Test Cases ***      baz         qux
        for index, token in enumerate(statement.data_tokens[:-1]):
            tokens.append(token)
            if self.min_width:
                separator = max(self.formatting_config.space_count, self.min_width - len(token.value)) * " "
            else:
                separator = (self.widths[index] - len(token.value) + self.formatting_config.space_count) * " "
            tokens.append(Token(Token.SEPARATOR, separator))
        tokens.append(statement.data_tokens[-1])
        tokens.append(statement.tokens[-1])  # eol
        statement.tokens = tokens
        return statement

    def align_statement(self, statement):
        tokens = []
        for line in statement.lines:
            strip_line = [t for t in line if t.type not in (Token.SEPARATOR, Token.EOL)]
            line_pos = 0
            exp_pos = 0
            widths = self.get_widths(statement)
            for token, width in zip(strip_line, widths):
                if self.min_width:
                    exp_pos += max(width + self.formatting_config.space_count, self.min_width)
                else:
                    exp_pos += width + self.formatting_config.space_count
                if self.test_name_len:
                    if self.name_line == statement.lineno:
                        exp_pos -= self.test_name_len
                    self.test_name_len = 0
                tokens.append(Token(Token.SEPARATOR, (exp_pos - line_pos) * " "))
                tokens.append(token)
                line_pos += len(token.value) + exp_pos - line_pos
            tokens.append(line[-1])
        statement.tokens = tokens

    def get_widths(self, statement):
        indent = self.indent
        if isinstance(statement, (ForHeader, End, IfHeader, ElseHeader, ElseIfHeader)):
            indent -= 1
        if not indent:
            return self.widths
        return [max(width, indent * self.formatting_config.space_count) for width in self.widths]

    def visit_SettingSection(self, node):  # noqa
        return node

    visit_VariableSection = visit_KeywordSection = visit_CommentSection = visit_SettingSection


class ColumnWidthCounter(ModelVisitor):
    def __init__(self, disablers):
        self.widths = []
        self.disablers = disablers
        self.test_name_lineno = -1
        self.any_one_line_test = False
        self.header_with_cols = False

    def visit_TestCaseSection(self, node):  # noqa
        self.generic_visit(node)
        if not self.header_with_cols and not self.any_one_line_test and self.widths:
            self.widths[0] = 0
        self.widths = [round_to_four(length) for length in self.widths]

    @skip_if_disabled
    def visit_Statement(self, statement):  # noqa
        if statement.type == Token.COMMENT:
            return
        if statement.type == Token.TESTCASE_HEADER:
            if len(statement.data_tokens) > 1:
                self.header_with_cols = True
                self._count_widths_from_statement(statement)
        elif statement.type == Token.TESTCASE_NAME:
            if self.widths:
                self.widths[0] = max(self.widths[0], len(statement.name))
            else:
                self.widths.append(len(statement.name))
            self.test_name_lineno = statement.lineno
        else:
            if self.test_name_lineno == statement.lineno:
                self.any_one_line_test = True
            if not isinstance(statement, (ForHeader, IfHeader, ElseHeader, ElseIfHeader, End)):
                self._count_widths_from_statement(statement, indent=1)

    def _count_widths_from_statement(self, statement, indent=0):
        for line in statement.lines:
            line = [t for t in line if t.type not in (Token.SEPARATOR, Token.EOL)]
            for index, token in enumerate(line, start=indent):
                if index < len(self.widths):
                    self.widths[index] = max(self.widths[index], len(token.value))
                else:
                    self.widths.append(len(token.value))



import re
from typing import Optional

from robot.api.parsing import Token

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.exceptions import InvalidParameterValueError
from robotidy.transformers import Transformer


class RenameTestCases(Transformer):
    r"""
    Enforce test case naming.

    Capitalize first letter of test case name, remove trailing dot and strip leading/trailing whitespace. If
    capitalize_each_word is true, will capitalize each word in test case name.

    It is also possible to configure `replace_pattern` parameter to find and replace regex pattern. Use `replace_to`
    to set replacement value. This configuration:

    ```
    robotidy --transform RenameTestCases -c RenameTestCases:replace_pattern=[A-Z]{3,}-\d{2,}:replace_to=foo
    ```

    will transform following code:

    ```robotframework
    *** Test Cases ***
    test ABC-123
        No Operation
    ```

    To:

    ```robotframework
    *** Test Cases ***
    Test foo
        No Operation
    ```

    ```
    robotidy --transform RenameTestCases -c RenameTestCases:capitalize_each_word=True
    ```

    will transform following code:

    ```robotframework
    *** Test Cases ***
    compare XML with json
        No Operation
    ```

    To:

    ```robotframework
    *** Test Cases ***
    Compare XML With Json
        No Operation
    ```
    """
    ENABLED = False

    def __init__(
        self,
        replace_pattern: Optional[str] = None,
        replace_to: Optional[str] = None,
        capitalize_each_word: bool = False,
    ):
        super().__init__()
        try:
            self.replace_pattern = re.compile(replace_pattern) if replace_pattern is not None else None
        except re.error as err:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "replace_pattern",
                replace_pattern,
                f"It should be a valid regex expression. Regex error: '{err.msg}'",
            )
        self.replace_to = "" if replace_to is None else replace_to
        self.capitalize_each_word = capitalize_each_word

    @skip_section_if_disabled
    def visit_TestCaseSection(self, node):  # noqa
        return self.generic_visit(node)

    @skip_if_disabled
    def visit_TestCaseName(self, node):  # noqa
        token = node.get_token(Token.TESTCASE_NAME)
        if token.value:
            if self.capitalize_each_word:
                token_value = " ".join(f"{word[0].upper()}{word[1:]}" for word in token.value.split(" "))
                token.value = token_value.lstrip()
            else:
                token.value = token.value[0].upper() + token.value[1:]
            if self.replace_pattern is not None:
                token.value = self.replace_pattern.sub(repl=self.replace_to, string=token.value)
            if token.value.endswith("."):
                token.value = token.value[:-1]
            token.value = token.value.strip()
        return node



try:
    from robot.api.parsing import InlineIfHeader, TryHeader
except ImportError:
    InlineIfHeader, TryHeader = None, None

from robotidy.disablers import skip_if_disabled
from robotidy.skip import Skip
from robotidy.transformers.aligners_core import AlignKeywordsTestsSection


class AlignKeywordsSection(AlignKeywordsTestsSection):
    """
    Align ``*** Keywords ***`` section to columns.

    Align keyword calls and settings into columns with predefined width. There are two possible alignment types
    (configurable via ``alignment_type``):
      - ``fixed`` (default): pad the tokens to the fixed width of the column
      - ``auto``: pad the tokens to the width of the longest token in the column

    Example output:
    ```robotframework
    *** Keywords ***
    Keyword
        ${var}        Create Resource       ${argument}       value
        Assert        value
        Multi
        ...           line
        ...           args
    ```

    Column widths can be configured via ``widths`` (default ``24``). It accepts comma separated list of column widths.

    Tokens that are longer than width of the column go into "overflow" state. It's possible to decide in this
    situation (by configuring ``handle_too_long``):
      - ``overflow`` (default): align token to the next column
      - ``compact_overflow``: try to fit next token between current (overflowed) token and the next column
      - ``ignore_rest``: ignore remaining tokens in the line
      - ``ignore_line``: ignore whole line

    It is possible to skip formatting on various types of the syntax (documentation, keyword calls with specific names
    or settings).
    """

    def __init__(
        self,
        widths: str = "",
        alignment_type: str = "fixed",
        handle_too_long: str = "overflow",
        compact_overflow_limit: int = 2,
        skip_documentation: str = "True",  # noqa - override skip_documentation from Skip
        skip: Skip = None,
    ):
        super().__init__(widths, alignment_type, handle_too_long, compact_overflow_limit, skip)

    @skip_if_disabled
    def visit_Keyword(self, node):  # noqa
        self.create_auto_widths_for_context(node)
        self.generic_visit(node)
        self.remove_auto_widths_for_context()
        return node

    def visit_TestCase(self, node):  # noqa
        return node



from robocorp_ls_core.constants import *
from robocorp_ls_core.options import USE_TIMEOUTS
from typing import Optional
from string import Template

DEFAULT_COMPLETIONS_TIMEOUT: int = 8
DEFAULT_COLLECT_DOCS_TIMEOUT: int = 40
DEFAULT_LIST_TESTS_TIMEOUT: int = 20
if not USE_TIMEOUTS:
    # A whole month of timeout seems good enough as a max.
    DEFAULT_COMPLETIONS_TIMEOUT = 60 * 60 * 24 * 30
    DEFAULT_COLLECT_DOCS_TIMEOUT = DEFAULT_COMPLETIONS_TIMEOUT
    DEFAULT_LIST_TESTS_TIMEOUT = DEFAULT_COMPLETIONS_TIMEOUT

# Robot Flow Explorer template used as wrapper for the React Application
# Substitute the $rfe_options & $rfe_data to render properly
DEFAULT_ROBOT_FLOW_EXPLORER_HTML_TEMPLATE: Template = Template(
    """
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no" />
    <title>Robot Flow Explorer</title>
    <link rel="icon" href="$rfe_favicon_path" />
    <script id="options" type="application/json">
      $rfe_options
    </script>
    <script id="data" type="application/json">
      $rfe_data
    </script>
  </head>
  <body>
    <noscript>Robot Visualization requires JavaScript</noscript>
    <div id="root"></div>
    <script defer="defer" src="$rfe_js_path"></script>
  </body>
</html>
"""
)

DEFAULT_ROBOT_FLOW_EXPLORER_OPTIONS: dict = {"theme": "light", "showCopyright": True}



from robot.api.parsing import Token

from robotidy.disablers import skip_section_if_disabled
from robotidy.transformers import Transformer


class NormalizeSectionHeaderName(Transformer):
    """
    Normalize section headers names.
    Robot Framework is quite flexible with the section header naming. Following lines are equal:

    ```robotframework
    *setting
    *** SETTINGS
    *** SettingS ***
    ```

    This transformer normalizes naming to follow ``*** SectionName ***`` format (with plural variant):

    ```robotframework
    *** Settings ***
    *** Keywords ***
    *** Test Cases ***
    *** Variables ***
    *** Comments ***
    ```

    Optional data after section header (for example data driven column names) is preserved.
    It is possible to upper case section header names by passing ``uppercase=True`` parameter:

    ```robotframework
    *** SETTINGS ***
    ```
    """

    def __init__(self, uppercase: bool = False):
        super().__init__()
        self.uppercase = uppercase

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    def visit_SectionHeader(self, node):  # noqa
        if node.name and "task" in node.name.lower():
            name = "*** Tasks ***"
        else:
            name = {
                Token.SETTING_HEADER: "*** Settings ***",
                Token.VARIABLE_HEADER: "*** Variables ***",
                Token.TESTCASE_HEADER: "*** Test Cases ***",
                Token.KEYWORD_HEADER: "*** Keywords ***",
                Token.COMMENT_HEADER: "*** Comments ***",
            }[node.type]
        if self.uppercase:
            name = name.upper()
        # we only modify header token value in order to preserver optional data driven testing column names
        node.data_tokens[0].value = name
        return node



# fmt: off
# Warning: Don't edit file (autogenerated from python -m dev codegen).

OPTION_ROBOT_LANGUAGE_SERVER_PYTHON = "robot.language-server.python"
OPTION_ROBOT_LANGUAGE_SERVER_ARGS = "robot.language-server.args"
OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT = "robot.language-server.tcp-port"
OPTION_ROBOT_PYTHON_EXECUTABLE = "robot.python.executable"
OPTION_ROBOT_PYTHON_ENV = "robot.python.env"
OPTION_ROBOT_VARIABLES = "robot.variables"
OPTION_ROBOT_LOAD_VARIABLES_FROM_ARGUMENTS_FILE = "robot.loadVariablesFromArgumentsFile"
OPTION_ROBOT_PYTHONPATH = "robot.pythonpath"
OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS = "robot.libraries.libdoc.needsArgs"
OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE = "robot.libraries.libdoc.preGenerate"
OPTION_ROBOT_CODE_FORMATTER = "robot.codeFormatter"
OPTION_ROBOT_FLOW_EXPLORER_THEME = "robot.flowExplorerTheme"
OPTION_ROBOT_LINT_ROBOCOP_ENABLED = "robot.lint.robocop.enabled"
OPTION_ROBOT_LINT_ENABLED = "robot.lint.enabled"
OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS = "robot.lint.undefinedKeywords"
OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES = "robot.lint.undefinedLibraries"
OPTION_ROBOT_LINT_UNDEFINED_RESOURCES = "robot.lint.undefinedResources"
OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS = "robot.lint.undefinedVariableImports"
OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS = "robot.lint.keywordCallArguments"
OPTION_ROBOT_LINT_KEYWORD_RESOLVES_TO_MULTIPLE_KEYWORDS = "robot.lint.keywordResolvesToMultipleKeywords"
OPTION_ROBOT_LINT_VARIABLES = "robot.lint.variables"
OPTION_ROBOT_LINT_IGNORE_VARIABLES = "robot.lint.ignoreVariables"
OPTION_ROBOT_LINT_IGNORE_ENVIRONMENT_VARIABLES = "robot.lint.ignoreEnvironmentVariables"
OPTION_ROBOT_LINT_UNUSED_KEYWORD = "robot.lint.unusedKeyword"
OPTION_ROBOT_COMPLETIONS_SECTION_HEADERS_FORM = "robot.completions.section_headers.form"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ENABLE = "robot.completions.keywordsNotImported.enable"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ADD_IMPORT = "robot.completions.keywordsNotImported.addImport"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_FORMAT = "robot.completions.keywords.format"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME = "robot.completions.keywords.prefixImportName"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME_IGNORE = "robot.completions.keywords.prefixImportNameIgnore"
OPTION_ROBOT_COMPLETIONS_KEYWORDS_ARGUMENTS_SEPARATOR = "robot.completions.keywords.argumentsSeparator"
OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS = "robot.workspaceSymbolsOnlyForOpenDocs"
OPTION_ROBOT_EDITOR_4SPACES_TAB = "robot.editor.4spacesTab"
OPTION_ROBOT_QUICK_FIX_KEYWORD_TEMPLATE = "robot.quickFix.keywordTemplate"
OPTION_ROBOT_CODE_LENS_ENABLE = "robot.codeLens.enable"
OPTION_ROBOT_CODE_LENS_RUN_ENABLE = "robot.codeLens.run.enable"
OPTION_ROBOT_CODE_LENS_INTERACTIVE_CONSOLE_ENABLE = "robot.codeLens.interactiveConsole.enable"
OPTION_ROBOT_INTERACTIVE_CONSOLE_ARGUMENTS = "robot.interactiveConsole.arguments"
OPTION_ROBOT_LANGUAGE = "robot.language"
OPTION_ROBOT_RUN_PEEK_ERROR_LEVEL = "robot.run.peekError.level"
OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_SUMMARY = "robot.run.peekError.showSummary"
OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_ERRORS_IN_CALLERS = "robot.run.peekError.showErrorsInCallers"

ALL_ROBOT_OPTIONS = frozenset(
    (
        OPTION_ROBOT_LANGUAGE_SERVER_PYTHON,
        OPTION_ROBOT_LANGUAGE_SERVER_ARGS,
        OPTION_ROBOT_LANGUAGE_SERVER_TCP_PORT,
        OPTION_ROBOT_PYTHON_EXECUTABLE,
        OPTION_ROBOT_PYTHON_ENV,
        OPTION_ROBOT_VARIABLES,
        OPTION_ROBOT_LOAD_VARIABLES_FROM_ARGUMENTS_FILE,
        OPTION_ROBOT_PYTHONPATH,
        OPTION_ROBOT_LIBRARIES_LIBDOC_NEEDS_ARGS,
        OPTION_ROBOT_LIBRARIES_LIBDOC_PRE_GENERATE,
        OPTION_ROBOT_CODE_FORMATTER,
        OPTION_ROBOT_FLOW_EXPLORER_THEME,
        OPTION_ROBOT_LINT_ROBOCOP_ENABLED,
        OPTION_ROBOT_LINT_ENABLED,
        OPTION_ROBOT_LINT_UNDEFINED_KEYWORDS,
        OPTION_ROBOT_LINT_UNDEFINED_LIBRARIES,
        OPTION_ROBOT_LINT_UNDEFINED_RESOURCES,
        OPTION_ROBOT_LINT_UNDEFINED_VARIABLE_IMPORTS,
        OPTION_ROBOT_LINT_KEYWORD_CALL_ARGUMENTS,
        OPTION_ROBOT_LINT_KEYWORD_RESOLVES_TO_MULTIPLE_KEYWORDS,
        OPTION_ROBOT_LINT_VARIABLES,
        OPTION_ROBOT_LINT_IGNORE_VARIABLES,
        OPTION_ROBOT_LINT_IGNORE_ENVIRONMENT_VARIABLES,
        OPTION_ROBOT_LINT_UNUSED_KEYWORD,
        OPTION_ROBOT_COMPLETIONS_SECTION_HEADERS_FORM,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ENABLE,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_NOT_IMPORTED_ADD_IMPORT,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_FORMAT,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_PREFIX_IMPORT_NAME_IGNORE,
        OPTION_ROBOT_COMPLETIONS_KEYWORDS_ARGUMENTS_SEPARATOR,
        OPTION_ROBOT_WORKSPACE_SYMBOLS_ONLY_FOR_OPEN_DOCS,
        OPTION_ROBOT_EDITOR_4SPACES_TAB,
        OPTION_ROBOT_QUICK_FIX_KEYWORD_TEMPLATE,
        OPTION_ROBOT_CODE_LENS_ENABLE,
        OPTION_ROBOT_CODE_LENS_RUN_ENABLE,
        OPTION_ROBOT_CODE_LENS_INTERACTIVE_CONSOLE_ENABLE,
        OPTION_ROBOT_INTERACTIVE_CONSOLE_ARGUMENTS,
        OPTION_ROBOT_LANGUAGE,
        OPTION_ROBOT_RUN_PEEK_ERROR_LEVEL,
        OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_SUMMARY,
        OPTION_ROBOT_RUN_PEEK_ERROR_SHOW_ERRORS_IN_CALLERS,
    )
)

# fmt: on
